/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * The findPlugin() functions use the ladspa SDK functions to search the
 * LADSPA_PATH. These functions take a function pointer and call this
 * function back. So, we write "wrapper" functions here that just return
 * a requested plugin.
 *
 * FIXME This actually results in a memory leak, since the plugin
 * libraries that are opened for searching are never closed. This,
 * however, shouldn't be a big problem, since dlopen() doesn't open
 * the same library more than once (it uses reference counting).
 ******************************************************************************/


#include <string.h>

#include <ladspa.h>

extern "C" {
#include <ladspa_sdk/utils.h>
}

#include <dladspa/find_plugin.h>

// The functions that actually interact with the LADSPA SDK functions;
// hidden inside an unnamed namespace, together with the data we need
// to pass to them (we can't pass them as parameters, because the functions
// are called indirectly by LADSPAPluginSearch, not directly by us).
namespace {

const LADSPA_Descriptor* descriptor;
unsigned long ID;
const char* label;

// These are called back by LADSPAPluginSearch, a function from the
// LADSPA SDK. To perform a single search, these functions may be called
// several times, one for each library file. Therefore, we may find the
// plugin and be called again afterwards. This means that we shouldn't
// alter the "descriptor" pointer if we don't find anything.
void
findByID (const char* libraryFileName, void* libraryHandle,
                    LADSPA_Descriptor_Function listDescriptor)
{
    int i = 0;
    const LADSPA_Descriptor* desc = 0;
    while ( (desc = listDescriptor (i++)) != 0) {
        if (desc->UniqueID == ID) {
            descriptor = desc;
            break;
        } // else, leave "descriptor" as it was
    }
};

void
findByName (const char* libraryFileName, void* libraryHandle,
                    LADSPA_Descriptor_Function listDescriptor)
{
    int i = 0;
    const LADSPA_Descriptor* desc = 0;
    while ( (desc = listDescriptor (i++)) != 0) {
        if (strcmp (desc->Label, label) == 0) {
            descriptor = desc;
            break;
        }  // else, leave "descriptor" as it was
    }
}

} // namespace

namespace DLADSPA {

// These are the functions the user can call
const LADSPA_Descriptor*
findPlugin (unsigned long requestedID)
{
    ID = requestedID;
    descriptor = 0;
    LADSPAPluginSearch (findByID);
    return descriptor;
};

const LADSPA_Descriptor*
findPlugin (const char* requestedLabel)
{
    label = requestedLabel;
    descriptor = 0;
    LADSPAPluginSearch (findByName);
    return descriptor;
};

} // namespace DLADSPA
