/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * This program handles requests from proxy plugins to create "real" plugins
 * on a remote machine. The proxy informs the host/port where is is willing to
 * receive the processed data; this program queries the nameserver in search
 * of servers available, and lists them to the user; the user chooses one
 * server and this program asks the server to instantiate the plugin, informing
 * the host/port of the proxy plugin. The server instantiates the plugin and
 * returns the host/port it will use to receive the data to be processed.
 * Finally, this program returns this host/port to the proxy plugin, which
 * can then use the remote plugin. When the proxy is deleted, it requests
 * to this program to delete the remote plugin. This program contacts the
 * server and asks for the plugin to be deleted.
 *
 * Requests to create/destroy plugins are sent from the proxy class to
 * the manager through TCP; these requests are interpreted by the MessageParser
 * class which calls the adequate methods on the RequestHandler class. The
 * RequestHandler class does the actual work and also presents the UI to the
 * user; to create new kinds of UI, we just need to create a derived class from
 * RequestHandler and override one method in it.
 ******************************************************************************/


#ifdef ORB_MICO
#include "plugin_factory/plugin_factory.h"
#include <mico/CosNaming.h>
#elif defined (ORB_OMNIORB)
#include "plugin_factory/plugin_factory.hh"
#endif

#include <signal.h>

#include <utility>
#include <vector>

#include <getopt.h>

#include <dladspa/constants.h>

#include <dlowlat/transport/ip/ipaddress.h>
#include <dlowlat/transport/ip/tcp/sockets.h>

#include <miscutils/corba_option_parser.h>

using namespace std;
using namespace DLowLat;
using namespace DLADSPA;


/******************************************************************************
 *                       Command-line parsing
 ******************************************************************************/

class OptionParser: public CORBAOptionParser
{
public:
    OptionParser (int& argc, char**& argv);
    OptionParser () { };
    DLowLat::IPAddress myAddress() const { return myAddress_; };
protected:
    void usage (string&);
private:
    static DLowLat::IPAddress myAddress_;
};

// Define storage for the static element in this class
IPAddress OptionParser::myAddress_;

OptionParser::OptionParser(int& argc, char**& argv) :
                    CORBAOptionParser (argc, argv)
{
    // Set the default host/port we bind to according to the environment;
    // if that's not set, use the built-in default. A command-line option
    // overrides both.
    char* managerEnv = getenv (MANAGER_ENV_NAME);
    if (managerEnv == 0) {
        myAddress_.host (MANAGER_DEFAULT_ADDRESS);
    } else {
        myAddress_.host (managerEnv);
    }

    // These are the options we recognize
    static struct option longOptions[] = {
        { "localport", 1, 0, 'p' },
        { "bindaddress", 1, 0, 'a' },
        { 0, 0, 0, 0 } // "Terminator"
    };

    while (1) {
        int option_index = 0;
        int c = getopt_long_only (argc, argv,
                 "p:a:", longOptions, &option_index);
        if (c == -1) // No more options
            break;
        switch (c) {
            case 'l':
                myAddress_.port(optarg);
                break;
            case 'a':
                myAddress_.host (optarg);
                break;
            default:
                string error = "Unknown option: -";
                error += c;
                usage(error);
        }
    }
};

void
OptionParser::usage (string& error)
{
    char* space = "        ";
    cerr << endl << endl;
    cerr << space << error << endl << endl << endl;
    
    cerr << space << "syntax: " << argv[0];
    cerr << " [--localport(p) portnumber] [--bindaddress(a) local_ip_address]";
    cerr << endl << endl << endl;
    exit (1);
};


/******************************************************************************
 *                            Handle Requests                                 *
 ******************************************************************************/

// Interpret incoming "events" (requests from the proxy) and call an
// appropriate handler. This is a kind of signal-based design.
// We could use the SigC++ library here, but our "signal handling" needs
// are too simple to justify that.
// This is the class that actually uses CORBA to locate plugin factories and
// asks them to create/delete plugins. It also is responsible for presenting
// the necessary UI to the user (using the Observer pattern or some form of
// MVC for the UI would be completely overkill here).

// This class should be a template, so that other transports than IP may
// be used. It's probably very easy do make this change.

class Manager {
public:
    Manager ();
    virtual ~Manager ();
    void operator () ();
private:
    inline void sendAnswer (const string& name, const IPAddress&);
    void createPlugin ();
    void deletePlugin ();
    // This actually creates the plugin
    IPAddress createPlugin (PluginFactory_var factory);

    // This method implements the UI of this class; windowSize is to be used
    // as a return value.
    virtual int showChooser (const vector<pair<string, PluginFactory_var> >&
                                                serverList);

    // Auxiliary methods
    vector<pair<string, PluginFactory_var> > findPluginFactories (void);
    PluginFactory_var findPluginFactory (const string& hostname);

    IPAddress myAddress;
    TcpServer sock;
    IPAddress receivedAddress;
    string spec;
    unsigned long sampleRate;
    string processorHostname;
    int windowSize;
};


/******************************************************************************
 *                                 main                                       *
 ******************************************************************************/

// These are the signal handler and the auxiliary variable that
// allow us to exit cleanly (without leaving open TCP connections).
bool killed = false;
void
clean_exit (int)
{
    if (killed) {
        // The previous signal didn't make us exit,
        // so let's use brute force.
        exit (0);
    } else {
        killed = true;
    }
};

// We perform and infinite loop, waiting for events (requests from some proxy)
// and handling them. We could use poll(), threads etc. here, but there's
// really no need to.

int
main (int argc, char** argv)
{
    struct sigaction sigact;
    sigact.sa_handler = clean_exit;
    sigaction (SIGTERM, &sigact, 0);
    sigaction (SIGHUP, &sigact, 0);
    sigaction (SIGINT, &sigact, 0);
    sigaction (SIGQUIT, &sigact, 0);

    OptionParser options (argc, argv);
    Manager manager;
    while (! killed) {
        manager ();
    }
};


/******************************************************************************
 *               Implementation of the Manager class
 ******************************************************************************/

Manager::Manager () : myAddress (OptionParser().myAddress()), sock (myAddress)
{
};

void
Manager::operator () ()
{
    sock.waitConnection ();
    if (! sock.isConnected()) {
        // This was caused by signal
        return;
    }
    string request;
    sock >> request;
    sampleRate = 0;
    if (request == "CREATE") {
        sock >> receivedAddress;
        spec = "";
        sock >> spec;
        sock >> sampleRate;
        createPlugin ();
    } else if (request == "DELETE") {
        processorHostname = "";
        sock >> processorHostname;
        sock >> receivedAddress;
        deletePlugin ();
    } else {
        // This should not happen!!!!
        cerr << "manager.cc, class Manager: Invalid request!" << endl;
        exit (1);
    }
};

inline void
Manager::sendAnswer (const string& name, const IPAddress& newAddress)
{
    sock << "CREATED";
    sock << name;
    sock << windowSize;
    sock << newAddress;
};

void
Manager::createPlugin ()
{
    vector<pair<string, PluginFactory_var> > serverList (findPluginFactories());
    int hostnum = showChooser (serverList);
    sendAnswer (serverList[hostnum].first,
                            createPlugin (serverList[hostnum].second));
};

// This function is specific for the UDP/IP transport
void
Manager::deletePlugin ()
{
    PluginFactory_var factory = findPluginFactory (processorHostname);
    struct IPAddressSpec pluginAddr;
    pluginAddr.host = receivedAddress.host().c_str();
    pluginAddr.port = receivedAddress.port();
    factory->deletePlugin (pluginAddr);
};

// This function is specific for the UDP/IP transport
IPAddress
Manager::createPlugin (PluginFactory_var factory)
{
    struct IPAddressSpec clientAddr, *serverAddr;
    clientAddr.host = receivedAddress.host().c_str();
    clientAddr.port = receivedAddress.port();
    serverAddr = factory->newPlugin
                    (clientAddr, spec.c_str(), sampleRate);
    IPAddress serverIP (string (serverAddr->host), serverAddr->port);
    return serverIP;
};

// The UI of this class; right now, a text interface
int
Manager::showChooser (const vector<pair<string, PluginFactory_var> >& serverList)
{
    cout << "Choose the host where the plugin is to run:" << endl;
    for (unsigned int i = 0; i < serverList.size(); ++i) {
        cout << i+1 << ". " << serverList[i].first << endl;
    }
    cout << "Host? " << flush;
    int hostnum;
    cin >> hostnum;
    --hostnum;
    cout << endl;
    cout << "Which is the window size to be used (0, 1 or 2)? " << flush;
    cin >> windowSize;
    return hostnum;
};

// Query the CORBA name service in search of the plugin factories available
vector<pair<string, PluginFactory_var> >
Manager::findPluginFactories ()
{
    /* Define the name of what we want (the DLADSPA naming context) */
    CosNaming::Name DLADSPAName;
    DLADSPAName.length(1);
    DLADSPAName[0].id = CORBA::string_dup("DLADSPA");
    DLADSPAName[0].kind = CORBA::string_dup("");

    /* query the nameserver */
    OptionParser options;
    CORBA::Object_ptr obj = options.inc()->resolve (DLADSPAName);
    CosNaming::NamingContext_var DLADSPAContext =
                                        CosNaming::NamingContext::_narrow (obj);
    CORBA::release (obj);

    // Retrieve all the factories registered in the naming context;
    // we'll assume the number of servers is not insanely large and
    // just keep references to all of them on a vector<>.
    CosNaming::BindingList_var list;
    CosNaming::BindingIterator_var it;
    vector<pair<string, PluginFactory_var> > factories;
    // 100 is a reasonable value, but anything would work
    DLADSPAContext->list (100, list, it);
    for (unsigned int i = 0; i < list->length(); ++i) {
        obj = DLADSPAContext->resolve (list[i].binding_name);
        PluginFactory_var factory = PluginFactory::_narrow (obj);
        CORBA::release (obj);
        factories.push_back (make_pair (string (list[i].binding_name[0].id), factory));
    }
    if (! CORBA::is_nil (it)) {
        while (it->next_n (100, list)) {
            for (unsigned int i = 0; i < list->length(); ++i) {
                obj = DLADSPAContext->resolve (list[i].binding_name);
                PluginFactory_var factory = PluginFactory::_narrow (obj);
                CORBA::release (obj);
                factories.push_back (make_pair
                                (string (list[i].binding_name[0].id), factory));
            }
        }
        it->destroy();
    }
    return factories;
};

// Find the factory that created a plugin
PluginFactory_var
Manager::findPluginFactory (const string& hostname)
{
    /* Define the name of what we want */
    CosNaming::Name serviceName;
    serviceName.length(2);
    serviceName[0].id = CORBA::string_dup("DLADSPA");
    serviceName[0].kind = CORBA::string_dup("");
    serviceName[1].id = CORBA::string_dup(hostname.c_str());
    serviceName[1].kind = CORBA::string_dup("pluginFactory");

    /* query the nameserver */
    OptionParser options;
    CORBA::Object_ptr obj = options.inc()->resolve (serviceName);
    PluginFactory_var factory = PluginFactory::_narrow (obj);
    CORBA::release (obj);
    return factory;
};

Manager::~Manager () { };