/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * Each port number of the virtual descriptor is mapped to a port of a
 * specific plugin instance. Therefore, there are two kinds of port
 * numbers in this code: the number of the virtual Descriptor's ports
 * and the number of the plugin's descriptor ports. This is a little
 * confusing, but I see no way to make it clearer.
 *
 * This class is too complex for what it does, the code is not very well
 * object-oriented and the relationship between MultiChannelPlugin and
 * MultiChannelPluginType is based on friendship, which is not great. I
 * don't see a way to clean up this code, though. I tried to compensate
 * for this with a lot of comments.
 *
 * FIXME! Right now, MultiChannelPlugin doesn't implement wetness; all
 * output ports have wetnesses of 100%. MultiChannelPlugin should also
 * implement run_adding and related stuff for plugins that don't have it.
 ******************************************************************************/

using namespace std;
#include <dladspa/multi_channel_plugin.h>

#include <dladspa/function_pointers.h>
#include <dladspa/find_plugin.h>

#include <sstream>
#include <set>

namespace DLADSPA {

/******************************************************************************
 *                       MultiChannelPluginType                               *
 ******************************************************************************/

// Most of the work is done inside the constructor:

MultiChannelPluginType::MultiChannelPluginType
            (const LADSPA_Descriptor* descriptor, unsigned long numChannels,
            const vector<PortOptions>& receivedControlOptions,
            const PortOptions& receivedWetnessOptions) :
            wetnessOptions (receivedWetnessOptions),
            pluginDescriptor (descriptor)
{
    realConstructor (numChannels, receivedControlOptions);
};

MultiChannelPluginType::MultiChannelPluginType
            (const char* pluginLabel, unsigned long numChannels,
            const vector<PortOptions>& receivedControlOptions,
            const PortOptions& receivedWetnessOptions) :
            wetnessOptions (receivedWetnessOptions)
{
    pluginDescriptor = findPlugin (pluginLabel);
    realConstructor (numChannels, receivedControlOptions);
};

MultiChannelPluginType::MultiChannelPluginType
            (unsigned long pluginID, unsigned long numChannels,
            const vector<PortOptions>& receivedControlOptions,
            const PortOptions& receivedWetnessOptions) :
            wetnessOptions (receivedWetnessOptions)
{
    pluginDescriptor = findPlugin (pluginID);
    realConstructor (numChannels, receivedControlOptions);
};

void
MultiChannelPluginType::realConstructor (unsigned long numChannels,
        const vector<PortOptions>& receivedControlOptions)
{
    // Figure out the relevant data on the wrapped plugin
    classifyPorts ();
    findExtraPorts ();
    mapControlOptions (receivedControlOptions);

    // How many instances of the plugin?
    this->numChannels = numChannels;
    findNumPlugins ();

    // Create the virtual ports
    mapVirtualPorts (pluginInputs);
    mapVirtualPorts (pluginOutputs);
    mapVirtualPorts (pluginControls);
    createWetnessPorts ();
    mapVirtualPorts (pluginExtraInputs);
    mapVirtualPorts (pluginExtraOutputs);

    // Create the virtual Descriptor
    createDescriptorMiscData ();
    setDescriptorFunctionPointers ();
    createDescriptorPortsData();
};

void
MultiChannelPluginType::classifyPorts ()
{
    // Classify the plugin's ports according to type and keep
    // the corresponding indexes in different collections (these
    // are indexes of the plugin descriptor's ports, not the
    // virtual descriptor's ports).
    for (unsigned long i = 0; i < pluginDescriptor->PortCount; ++i) {
        if (LADSPA_IS_PORT_AUDIO (pluginDescriptor->PortDescriptors[i])) {
            if (LADSPA_IS_PORT_INPUT (pluginDescriptor->PortDescriptors[i])) {
                pluginInputs.push_back (i);
            } else {
                pluginOutputs.push_back (i);
            }
        } else {
            pluginControls.push_back (i);
        }
    }
};

void
MultiChannelPluginType::findExtraPorts ()
{
    // Each plugin has a set of inputs, outputs, extra inputs
    // and extra outputs; separate the "normal" and "extra"
    // ports indexes here. (these are indexes of the plugin
    // descriptor's ports, not the virtual descriptor's ports).
    unsigned long numExtraInputs, numExtraOutputs;
    if (pluginInputs.size() == 0 || pluginOutputs.size() == 0) {
        // This is a "special" plugin; since there are either no inputs
        // or no outputs, there's no sense in thinking about "extra"
        // ports.
        numExtraInputs = numExtraOutputs = 0;
    } else if (pluginInputs.size() < pluginOutputs.size()) {
        numExtraInputs = 0;
        numExtraOutputs = pluginOutputs.size() - pluginInputs.size();
    } else {
        numExtraOutputs = 0;
        // This may be zero.
        numExtraInputs = pluginInputs.size() - pluginOutputs.size();
    }

    vector<unsigned long>::iterator it;

    it = pluginInputs.begin();
    it += pluginInputs.size() - numExtraInputs;
    pluginExtraInputs.insert (pluginExtraInputs.end(), it, pluginInputs.end());
    pluginInputs.erase (it, pluginInputs.end());

    it = pluginOutputs.begin();
    it += pluginOutputs.size() - numExtraOutputs;
    pluginExtraOutputs.insert (pluginExtraOutputs.end(), it, pluginOutputs.end());
    pluginOutputs.erase (it, pluginOutputs.end());
};

void
MultiChannelPluginType::mapControlOptions
                 (const vector<PortOptions>& receivedControlOptions)
{
    // Each plugin control port has a "locked" and an "enabled"
    // status associated with it. It also has a collection of default
    // values that should be applied to that port on each plugin instance.
    // Here, we create the mapping from the plugin port index to the
    // PortOptions struct that defines these attributes (these are indexes
    // of the plugin descriptor's ports, not the virtual descriptor's ports).
    vector<unsigned long>::iterator it;
    for (it = pluginControls.begin(); it != pluginControls.end(); ++it) {
        controlOptions.insert (make_pair
                        (*it, receivedControlOptions[controlOptions.size()]));
    }
};

void
MultiChannelPluginType::findNumPlugins () 
{
    // How many plugin instances do we need to have the
    // correct number of channels?
    if (pluginOutputs.size() == 0) {
        // This is probably a synthesizer that only has outputs.
        // Let's try to deal with it
        if (numChannels % pluginOutputs.size() != 0) {
            // This shouldn't happen, as jack-rack doesn't create
            // such a chain
            exit (1);
        } else {
            numPlugins = numChannels / pluginOutputs.size();
        }
    } else if (pluginInputs.size () == 0) {
        // This is probably a meter plugin or something; let's try
        // to deal with it
        if (numChannels % pluginInputs.size() != 0) {
            // This shouldn't happen, as jack-rack doesn't create
            // such a chain
            exit (1);
        } else {
            numPlugins = numChannels / pluginInputs.size();
        }
    } else if (pluginInputs.size() < pluginOutputs.size()) {
        if ((numChannels % pluginInputs.size()) != 0) {
            // This shouldn't happen, as jack-rack doesn't create
            // such a chain
            exit (1);
        } else {
            numPlugins = numChannels / pluginInputs.size();
        }
    } else {
        if ((numChannels % pluginOutputs.size()) != 0) {
            // This shouldn't happen, as jack-rack doesn't create
            // such a chain
            exit (1);
        } else {
            numPlugins = numChannels / pluginOutputs.size();
        }
    }
};

void
MultiChannelPluginType::mapVirtualPorts (const vector<unsigned long>& ports)
{
    // This method creates the mapping from virtual ports on a
    // MultiChannelPlugin of the type we represent to real ports
    // on instances of the plugin that are held inside the
    // MultiChannelPlugin.

    // This is the collection of locked ports that have already
    // been inserted once and that, therefore, shouldn't be
    // inserted again.
    set<unsigned long> processed;
    for (unsigned long i = 0; i < numPlugins; ++i) {
        vector<unsigned long>::const_iterator it;
        for (it = ports.begin(); it != ports.end(); ++it) {
            if (processed.find (*it) != processed.end() || ! isEnabled (*it)) {
                // This port shouldn't be mapped
                continue;
            } else {
                virtualPorts.push_back (make_pair(i,*it));
                if (isLocked (*it)) {
                    processed.insert (*it);
                }
            }
        }
    }
};

void
MultiChannelPluginType::createWetnessPorts ()
{
    // We will create new "fake" control ports on the plugin;
    // each of these additional control ports is the wetness
    // control of a corresponding output port (not an extra
    // port!).
    if (! wetnessOptions.enabled) {
        return;
    }
    if (wetnessOptions.locked) {
        // Only one wetness virtual port; let's map it to the
        // first output, although it controls all of them.
        virtualPorts.push_back (make_pair (0, pluginOutputs[0]));
    } else {
        for (unsigned long i = 0; i < numPlugins; ++i) {
            for (unsigned long j = 0; j < pluginOutputs.size(); ++j) {
                virtualPorts.push_back (make_pair(i, j));
                wetnessPortNumbers.push_back (virtualPorts.size() -1);
            }
        }
    }
};

void
MultiChannelPluginType::createDescriptorMiscData ()
{
    // FIXME!!!!! We need a way to deal with UniqueID cleanly!
    virtualDescriptor.UniqueID = 0;

    // New label and name, based on old.
    if (numPlugins > 1) {
        ostringstream s;
        s << "multi_" << pluginDescriptor->Label << "_" << numPlugins << "_times";
        virtualDescriptor.Label = strdup (s.str().c_str());
        s.str("");
        s << "Multi " << pluginDescriptor->Name;
        virtualDescriptor.Name = strdup (s.str().c_str());
    } else {
        virtualDescriptor.Label = strdup (pluginDescriptor->Label);
        virtualDescriptor.Name = strdup (pluginDescriptor->Name);
    }

    // Copy most of the original's data; we don't do a deep copy here
    // bacause the plugin descriptor must continue to be valid as long
    // as this object exists.
    virtualDescriptor.Properties = pluginDescriptor->Properties;
    virtualDescriptor.Maker = pluginDescriptor->Maker;
    virtualDescriptor.Copyright = pluginDescriptor->Copyright;

    // When we instantiate the plugin, we need to access this instance
    // of the MultiChannelPluginType to know what the plugin to be
    // instantiated looks like.
    virtualDescriptor.ImplementationData = this;
};

void
MultiChannelPluginType::setDescriptorFunctionPointers ()
{
    virtualDescriptor.instantiate = ::instantiate<MultiChannelPlugin>;
    virtualDescriptor.connect_port = ::connect_port<MultiChannelPlugin>;
    virtualDescriptor.activate = ::activate<MultiChannelPlugin>;
    virtualDescriptor.run = ::run<MultiChannelPlugin>;
    if (pluginDescriptor->run_adding != 0) {
        virtualDescriptor.run_adding = ::run_adding<MultiChannelPlugin>;
        virtualDescriptor.set_run_adding_gain =
                            ::set_run_adding_gain<MultiChannelPlugin>;
    } else {
        virtualDescriptor.run_adding = 0;
        virtualDescriptor.set_run_adding_gain = 0;
    }
    virtualDescriptor.deactivate = ::deactivate<MultiChannelPlugin>;
    virtualDescriptor.cleanup = ::cleanup<MultiChannelPlugin>;
};

void
MultiChannelPluginType::createDescriptorPortsData ()
{
    // Now create the virtual ports
    unsigned long numPorts = virtualPorts.size();
    virtualDescriptor.PortCount = numPorts;
    virtualDescriptor.PortDescriptors =
                new LADSPA_PortDescriptor[numPorts];
    virtualDescriptor.PortNames = new const char* [numPorts];
    virtualDescriptor.PortRangeHints =
                new LADSPA_PortRangeHint[numPorts];

    // Define each port's info: name, hints etc. For all except
    // the wetness control ports, just copy the data from the
    // real plugin.
    for (unsigned long virtualPort = 0; virtualPort < virtualPorts.size();
                                                        ++virtualPort) {
        unsigned long pluginPort, instance;
        instance = virtualPorts[virtualPort].first;
        pluginPort = virtualPorts[virtualPort].second;
        if (isWetness(virtualPort)) {
            createWetnessPortData (virtualPort, pluginPort, instance);
        } else {
            copyPortData (virtualPort, pluginPort, instance);
        }
    }
};

void
MultiChannelPluginType::createWetnessPortData (unsigned long virtualPort,
                            unsigned long pluginPort, unsigned long instance)
{
    const_cast<LADSPA_PortDescriptor*>
                    (virtualDescriptor.PortDescriptors)[virtualPort] =
                    LADSPA_PORT_CONTROL | LADSPA_PORT_INPUT;
    const_cast<LADSPA_PortRangeHint*>
        (virtualDescriptor.PortRangeHints)[virtualPort].HintDescriptor =
            LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE |
            LADSPA_HINT_LOGARITHMIC | LADSPA_HINT_DEFAULT_MAXIMUM;
    const_cast<LADSPA_PortRangeHint*>
        (virtualDescriptor.PortRangeHints)[virtualPort].LowerBound = 0.0;
    const_cast<LADSPA_PortRangeHint*>
        (virtualDescriptor.PortRangeHints)[virtualPort].UpperBound = 1.0;

    ostringstream name;
    name << "wetness";
    if (numPlugins > 1 && ! wetnessOptions.locked) {
        name << "_" << instance;
    }
    const_cast<char**> (virtualDescriptor.PortNames)[virtualPort] =
        strdup (name.str().c_str());
};

void
MultiChannelPluginType::copyPortData (unsigned long virtualPort,
                            unsigned long pluginPort, unsigned long instance)
{
    // Just copy the original data
    const_cast<LADSPA_PortDescriptor*>
        (virtualDescriptor.PortDescriptors)[virtualPort] =
                    pluginDescriptor->PortDescriptors[pluginPort];
    const_cast<LADSPA_PortRangeHint*>
            (virtualDescriptor.PortRangeHints)[virtualPort] =
                    pluginDescriptor->PortRangeHints[pluginPort];
    if (numPlugins > 1 && ! isLocked (pluginPort)) {
        ostringstream name;
        name << pluginDescriptor->PortNames[pluginPort] << "_" << instance;
        const_cast<char**>
            (virtualDescriptor.PortNames)[virtualPort] =
            strdup (name.str().c_str());
    } else {
        const_cast<char**>
            (virtualDescriptor.PortNames)[virtualPort] =
            strdup (pluginDescriptor->PortNames[pluginPort]);
    }
};

bool
MultiChannelPluginType::isEnabled (unsigned long portNum) const
{
    map<unsigned long, PortOptions>::const_iterator it;
    if ((it = controlOptions.find (portNum)) != controlOptions.end()) {
        return it->second.enabled;
    } else {
        // This is an audio port
        return true;
    }
};

bool
MultiChannelPluginType::isLocked (unsigned long portNum) const
{
    map<unsigned long, PortOptions>::const_iterator it;
    if ((it = controlOptions.find (portNum)) != controlOptions.end()) {
        return it->second.locked;
    } else {
        // This is an audio port
        return false;
    }
};

MultiChannelPluginType::~MultiChannelPluginType ()
{
    free (const_cast<char*>(virtualDescriptor.Label));
    free (const_cast<char*>(virtualDescriptor.Name));
    for (unsigned long i = 0; i < virtualDescriptor.PortCount; ++i) {
        free (const_cast<char**>(virtualDescriptor.PortNames)[i]);
    }
    delete[] virtualDescriptor.PortDescriptors;
    delete[] virtualDescriptor.PortNames;
    delete[] virtualDescriptor.PortRangeHints;
};


/******************************************************************************
 *                            MultiChannelPlugin                              *
 ******************************************************************************/

MultiChannelPlugin::MultiChannelPlugin (const LADSPA_Descriptor* descriptor,
                    unsigned long sampleRate)
{
    myType = static_cast<MultiChannelPluginType*>
                        (descriptor->ImplementationData);

    // Create the plugin instances
    for (unsigned long i = 0; i < myType->numPlugins; ++i) {
        plugins.push_back (myType->pluginDescriptor->instantiate
                            (myType->pluginDescriptor, sampleRate));
    }

    // Connect the control and wetness ports to the default values.
    // First, the control ports...
    for (vector<unsigned long>::iterator it = myType->pluginControls.begin();
                            it != myType->pluginControls.end(); ++it) {
        vector<LADSPA_Data>& values (myType->controlOptions[*it].values);
        for (unsigned long i = 0; i < myType->numPlugins; ++i) {
            myType->pluginDescriptor->connect_port
                (plugins[i], *it, &(values[i]));
        }
    }

    // Then, the wetness ports.
    wetnessPortAddresses.resize (myType->wetnessPortNumbers.size());
    vector<LADSPA_Data>& values (myType->wetnessOptions.values);
    for (unsigned long i = 0; i < values.size(); ++i) {
        wetnessPortAddresses.push_back (&(values[i]));
    }
};

MultiChannelPlugin::~MultiChannelPlugin ()
{
    if (myType->pluginDescriptor->cleanup) {
        for (unsigned long i = 0; i < myType->numPlugins; ++i) {
            myType->pluginDescriptor->cleanup (plugins[i]);
        }
    }
};

void
MultiChannelPlugin::connect_port (unsigned long port, LADSPA_Data* ptr)
{
    unsigned long instance = myType->virtualPorts[port].first;
    unsigned long pluginPort = myType->virtualPorts[port].second;

    if (myType->isWetness (port)) {
        if (myType->wetnessOptions.locked) {
            // Connect all wetness pointers to the received address
            for (unsigned long i = 0; i < wetnessPortAddresses.size(); ++i) {
                wetnessPortAddresses[i] = ptr;
            }
        } else {
            // Connect a single wetness port
            wetnessPortAddresses[port - myType->wetnessPortNumbers[0]] = ptr;
        }
    } else {
        // This is not a wetness port
        if (myType->isLocked(pluginPort)) {
            // This is a locked control port;
            // connect this port to all instances
            for (unsigned long i = 0; i < myType->numPlugins; ++i) {
                myType->pluginDescriptor->connect_port
                                                (plugins[i], pluginPort, ptr);
            }
        } else {
            // This is an ordinary port, just connect it.
            myType->pluginDescriptor->connect_port
                                        (plugins[instance], pluginPort, ptr);
        }
    }
};

} // namespace
