#ifndef DLADSPA_MULTICHAN_PLUGIN_H
#define DLADSPA_MULTICHAN_PLUGIN_H

/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * MultiChannelPlugin wraps several "parallel" instances of a single plugin.
 * These instances are used to mimic a plugin with a specific number of
 * channels. For instance, if we want a stereo amplifier but the only
 * amplifier LADSPA plugin available is mono, we instantiate two of them
 * inside this class and offer 2 ins and 2 outs to the client.
 *
 * MultiChannelPluginType receives the data on the kind of MultiChannelPlugin
 * we want to create on the constructor and creates a LADSPA_Descriptor
 * for this MultiChannelPlugin as well as some data needed to actually
 * instantiate the MultiChannelPlugin. On the created LADSPA_Descriptor,
 * MultiChannelPluginType puts a reference to itself in ImplementationData;
 * this is used when the client calls instantiate on the plugin (it's
 * the only way MultiChannelPlugin "knows" how it should be created). This
 * means that, as long as the LADSPA_Descriptor is in use, the
 * MultiChannelPluginType must remain a valid object.
 *
 * If the number of inputs of the wrapped plugin is different from the
 * number of outputs, the "extra" ports appear as inputs or outputs
 * after the "normal" ins/outs; therefore, if the client requested
 * a wrapper with 4 channels, it should "know" that inputs or output
 * ports beyond the 4th are "extras". These ports are called "aux"
 * ports in jack-rack.
 * 
 * Each control port may or may not be locked. If a control port is locked,
 * that port in all the plugin instances is exposed to the client as a
 * single port (that is, connect_port will map the port of all the instances
 * to the same address). If a port is not locked, each instance's port will
 * be exposed to the client separately. In the case of the example above,
 * the "gain" port may be locked (changing the gain changes both channels
 * simultaneously) or not (there's a gain control for each channel).
 *
 * A control port may also be enabled or disabled; if the port is disabled,
 * it is not exposed to the client at all: instead, the port of each instance
 * will be mapped internally to a default value.
 *
 * Finally, the audio outputs have "wetness" controls that may also be locked
 * or not and that may be enabled or not.
 *
 * It would be nice if we could inform the client of the default value of
 * a port: that is, even enabled ports could make use of the control
 * values supplied by the client as defaults. This is not possible, though,
 * because LADSPA offers no way of offering these default values to the
 * LADSPA client. Maybe this will be addressed on a future version of LADSPA.
 *
 * Jack-rack doesn't offer a way to define the "enabled/disabled" status of
 * a port; this needs to be implemented in jack-rack.
 ******************************************************************************/

#include <ladspa_sdk/ladspa.h>

#include <map>
#include <vector>
#include <functional>
#include <algorithm>

namespace DLADSPA {

// The default values and locked/enabled status of the wetness and
// control ports. Each instance of this struct describes the wetnesses
// or a control port on the actual plugins inside MultiChannelPlugin,
// not virtual ports. The client supplies structs of this type to
// the constructor of the MultiChannelPluginType class.
struct PortOptions {
    // Do all instances of this port share the same value?
    bool locked;
    // Should this appear as a LADSPA control port?
    bool enabled;
    // These are the default values for the wetnesses and control
    // ports.
    std::vector<LADSPA_Data> values;
};

// This is a description of what a MultiChannelPlugin instance should
// look like: it has a virtual LADSPA_Descriptor, the LADSPA_Descriptor
// of the wrapped plugin and the data relevant to each plugin port.
// An instance of this class is stored in the corresponding virtual
// descriptor "ImplementationData" pointer and is used when we call
// "instantiate".
class MultiChannelPluginType {
friend class MultiChannelPlugin;
public:
    MultiChannelPluginType (const LADSPA_Descriptor*,
                unsigned long numChannels,
                const std::vector<PortOptions>& controlOptions,
                const PortOptions& wetnessOptions);
    MultiChannelPluginType (const char* pluginLabel,
                unsigned long numChannels,
                const std::vector<PortOptions>& controlOptions,
                const PortOptions& wetnessOptions);
    MultiChannelPluginType (unsigned long pluginID,
                unsigned long numChannels,
                const std::vector<PortOptions>& controlOptions,
                const PortOptions& wetnessOptions);
    ~MultiChannelPluginType ();
    inline const LADSPA_Descriptor* descriptor () const;
private:
    // These are accessed by the friend "MultiChannelPlugin" class:

    // All the port numbers below refer to the plugin descriptor's
    // ports, not the virtual descriptor's ports.
    std::vector<unsigned long> pluginInputs, pluginOutputs, pluginControls,
                    pluginExtraInputs, pluginExtraOutputs;
    // This links the control port numbers to the respective
    // PortOptions.
    std::map<unsigned long, PortOptions> controlOptions;
    // Is this control port enabled?
    bool isEnabled (unsigned long portNum) const;
    // Is this control port locked?
    bool isLocked (unsigned long portNum) const;

    // Wetness ports are virtual ports
    std::vector<unsigned long> wetnessPortNumbers;
    PortOptions wetnessOptions;
    // Is this (virtual) port a wetness port?
    inline bool isWetness (unsigned long portnum) const;

    const LADSPA_Descriptor *pluginDescriptor;
    LADSPA_Descriptor virtualDescriptor;
    unsigned long numChannels, numPlugins;

    // This maps the virtual port numbers to the corresponding plugin
    // instance and its port. The vector index is the virtual port
    // number and the pair parameters are the plugin instance number
    // and the plugin's port number. If the virtualPort is a wetness
    // port, the referenced port is the controlled port.
    std::vector<std::pair<unsigned long, unsigned long> > virtualPorts;

    // These are conceptually parts of the constructor, and
    // are really private (that is, the friend class doesn't
    // access them):
    void realConstructor (unsigned long numChannels,
                            const std::vector<PortOptions>& controlOptions);
    void classifyPorts ();
    void findExtraPorts ();
    void mapControlOptions (const std::vector<PortOptions>&);
    void findNumPlugins ();
    void mapVirtualPorts (const std::vector<unsigned long>& ports);
    void createWetnessPorts ();
    // These are the methods that create the virtual Descriptor:
    void createDescriptorMiscData ();
    void setDescriptorFunctionPointers ();
    void createDescriptorPortsData ();
    void createWetnessPortData (unsigned long virtualPort,
                            unsigned long pluginPort, unsigned long instance);
    void copyPortData (unsigned long virtualPort, unsigned long pluginPort,
                                                unsigned long instance);
};

class MultiChannelPlugin {
public:
    MultiChannelPlugin (const LADSPA_Descriptor*, unsigned long sampleRate);
    ~MultiChannelPlugin ();
    inline void activate ();
    inline void deactivate ();
    void connect_port (unsigned long port, LADSPA_Data* ptr);
    inline void run (unsigned long sampleCount);
    inline void run_adding (unsigned long sampleCount);
    inline void set_run_adding_gain (LADSPA_Data gain);
    inline void cleanup ();
private:
    std::vector<LADSPA_Handle> plugins;
    MultiChannelPluginType* myType;
    // The wetness ports, differently from the other virtual ports, don't
    // refer to a real port on a plugin instance. Therefore, we need to keep
    // the pointers to these ports inside this class. So, we keep a
    // vector<LADSPA_Data*> with the pointers to the wetness ports.
    std::vector<LADSPA_Data*> wetnessPortAddresses;
};


/******************************************************************************
 *                                Inlines                                     *
 ******************************************************************************/

inline const LADSPA_Descriptor*
MultiChannelPluginType::descriptor () const
{
    return &virtualDescriptor;
};

inline bool
MultiChannelPluginType::isWetness (unsigned long portNum) const
{
    // The wetness ports are stored sequentially
    if (wetnessPortNumbers.empty()) {
        return false;
    } else {
        return (portNum >= wetnessPortNumbers.front() &&
            portNum <= wetnessPortNumbers.back());
    }
};

inline void
MultiChannelPlugin::activate ()
{
    if (myType->pluginDescriptor->activate) {
        for_each (plugins.begin(), plugins.end(),
                myType->pluginDescriptor->activate);
    }
};

inline void
MultiChannelPlugin::deactivate ()
{
    if (myType->pluginDescriptor->deactivate) {
        for_each (plugins.begin(), plugins.end(),
                myType->pluginDescriptor->deactivate);
    }
};

inline void
MultiChannelPlugin::run (unsigned long sampleCount)
{
    std::for_each (plugins.begin(), plugins.end(), std::bind2nd
                (std::ptr_fun (myType->pluginDescriptor->run), sampleCount));
};

inline void
MultiChannelPlugin::run_adding (unsigned long sampleCount)
{
    // FIXME! Implement run_adding if the plugin doesn't offer it.
    // To do this, we copy the output buffer to a temporary buffer,
    // call the plugins to do the processing and copy the temporary
    // buffer back, with scaling and summing with the processed result.
    if (myType->pluginDescriptor->run_adding) {
        std::for_each (plugins.begin(), plugins.end(), std::bind2nd
            (std::ptr_fun (myType->pluginDescriptor->run_adding), sampleCount));
    }
};

inline void
MultiChannelPlugin::set_run_adding_gain (LADSPA_Data gain)
{
    // FIXME! Implement this for plugins that don't offer run_adding.
    if (myType->pluginDescriptor->set_run_adding_gain) {
        std::for_each (plugins.begin(), plugins.end(), std::bind2nd
            (std::ptr_fun (myType->pluginDescriptor->set_run_adding_gain), gain));
    }
};

inline void
MultiChannelPlugin::cleanup ()
{
    // If the plugin doesn't offer this function, there's a memory leak.
    if (myType->pluginDescriptor->cleanup) {
        std::for_each (plugins.begin(), plugins.end(),
                std::ptr_fun (myType->pluginDescriptor->cleanup));
    }
};

} // namespace
#endif // ifndef DLADSPA_MULTICHAN_PLUGIN_H
