/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * When we instantiate a group of plugins "inside" a PluginChain, we need to
 * connect the outputs of each plugin to the inputs of the next one, so that
 * the client only "sees" the inputs of the first and the outputs of the last
 * plugins. We keep the sequence of interconnected ports that correspond to
 * a single channel on the ChannelPortsSequence vector.
 * 
 *  While generally it may be better to use the output buffers defined by the
 * client to interconnect the intermediate plugins (so that all processing is
 * done in place), some plugins are INPLACE_BROKEN. If this is the case, we
 * need to make use of an intermediate buffer, that is hidden from the client
 * inside PluginChain. These buffers are kept in the buffers vector.
 * 
 * When the client calls connect_port, it may make us start doing inplace
 * processing where we weren't doing it before, or it may make us stop
 * doing inplace processing where we were doing it before. Therefore, we
 * need to know, when connect_port is called, what is the address of the
 * "other" side of the port. to do this, we keep a record of all the current
 * pointers assigned to a plugin port in the inputAddresses and outputAddresses
 * vectors and reconnect all relevant ports when a connect_port request is
 * received.
 * 
 * Just like MultiChannelPlugin, this class is too complex for what it does
 * and the code isn't pretty. Still, I see no way to make it better.
 ******************************************************************************/

#include <iostream>
#include <sstream>

using namespace std;

extern "C" {
#include <stdlib.h>
}

extern "C" {
#include <libxml/tree.h>
}

#include <dladspa/plugin_chain.h>
#include <dladspa/find_plugin.h>
#include <dladspa/function_pointers.h>
#include <dladspa/multi_channel_plugin.h>

namespace DLADSPA {


/******************************************************************************
 *                              PluginChain                                   *
 ******************************************************************************/

void
PluginChain::connect_port (unsigned long port, LADSPA_Data* ptr)
{
    // The virtualPorts are organized as: all inputs, all outputs,
    // all controls, all extra inputs and all extra outputs.
    if (port >= 2 * myType->numChannels) {
        // This is a control or an extra input/output port;
        // we just need to connect it.
        unsigned long plugin = myType->virtualPorts[port].first;
        unsigned long pluginPort = myType->virtualPorts[port].second;
        myType->pluginTypes[plugin]->descriptor()->connect_port
                                (plugins[plugin], pluginPort, ptr);
    } else {
        // This is an audio port that belongs to a channel; we
        // need to reconnect all the ports that belong to the
        // channel.
        unsigned long channel = 0;
        if (port < myType->numChannels) {
            // This is an input port
            channel = port;
            inputAddresses[channel] = ptr;
        } else {
            // This is an output port
            channel = port - myType->numChannels;
            outputAddresses[channel] = ptr;
        }
        connectChannel (channel);
    }
};

void
PluginChain::connectChannel (unsigned long channel)
{
    if (inputAddresses[channel] == 0 || outputAddresses[channel] == 0) {
        // We can't connect everything if we don't have both
        // the input and the output port. FIXME This obviously
        // doesn't work for plugins that don't have either
        // inputs or outputs, like oscillators.
        return;
    }

    // We need to switch between the real output buffer and the
    // temp buffer when we face a plugin that is inPlaceBroken;
    // this array and the chooser variable help with that.
    // If there's an odd number of inPlaceBroken plugins, we
    // would switch between the output buffer and the temp
    // buffer an odd number of times; we need to do that an
    // even number of times, so in that case we perform an
    // additional switch with the first plugin, which is *not*
    // inPlaceBroken. If, OTOH, the number is odd but the first
    // plugin *is* inPlaceBroken, we just pretend it isn't and
    // everything goes fine.
    LADSPA_Data* tmpbuf[2];
    tmpbuf[0] = outputAddresses[channel];
    tmpbuf[1] = buffers[channel];
    unsigned long chooser = 0;

    unsigned long pluginInput, pluginOutput;

    // Process the first plugin
    pluginInput = myType->channelPortsSequence[channel][0].first;
    pluginOutput = myType->channelPortsSequence[channel][0].second;
    // Connect the input
    myType->pluginTypes[0]->descriptor()->connect_port
            (plugins[0], pluginInput, inputAddresses[channel]);
    if ((myType->numInPlaceBroken % 2 != 0) && (! LADSPA_IS_INPLACE_BROKEN
                        (myType->pluginTypes[0]->descriptor()->Properties))) {
        chooser = 1;
    }
    // connect the output of the plugin
    if (plugins.size() == 1) {
        // This is the last plugin; connect to the output
        myType->pluginTypes[0]->descriptor()->connect_port
                        (plugins[0], pluginOutput, outputAddresses[channel]);
    } else {
        myType->pluginTypes[0]->descriptor()->connect_port
                        (plugins[0], pluginOutput, tmpbuf[chooser % 2]);
    }

    // Now the other plugins (starting from 1), switching buffers if
    // the plugin is inPlaceBroken.
    for (unsigned long i = 1; i < plugins.size(); ++i) {
        pluginInput = myType->channelPortsSequence[channel][i].first;
        pluginOutput = myType->channelPortsSequence[channel][i].second;

        myType->pluginTypes[i]->descriptor()->connect_port
                    (plugins[i], pluginInput, tmpbuf[chooser % 2]);
        if (LADSPA_IS_INPLACE_BROKEN
                (myType->pluginTypes[i]->descriptor()->Properties)) {
            ++chooser;
        }
        if (i == plugins.size() -1) {
            // This is the last plugin; connect to the output
            myType->pluginTypes[i]->descriptor()->connect_port
                        (plugins[i], pluginOutput, outputAddresses[channel]);
        } else {
            myType->pluginTypes[i]->descriptor()->connect_port
                            (plugins[i], pluginOutput, tmpbuf[chooser % 2]);
        }
    }
};

PluginChain::~PluginChain ()
{
    for (unsigned long i = 0; i < plugins.size(); ++i) {
        myType->pluginTypes[i]->descriptor()->cleanup(plugins[i]);
    }

    for (unsigned long i = 0; i < buffers.size(); ++i) {
        delete buffers[i];
    }
};

PluginChain::PluginChain (const LADSPA_Descriptor* descriptor,
                                                    unsigned long sampleRate)
{
    // FIXME We need to deal with MultiChannelPlugins that either have
    // no inputs or no outputs; these only make sense on the beginning
    // or end of the chain respectively, in which case the PluginChain
    // also doesn't have the corresponding ports.
    myType = static_cast<PluginChainType*> (descriptor->ImplementationData);
    inputAddresses.resize (myType->numChannels, 0);
    outputAddresses.resize (myType->numChannels, 0);
    buffers.resize (myType->numChannels, 0);
    for (unsigned long i = 0; i < myType->pluginTypes.size(); ++i) {
        plugins.push_back (myType->pluginTypes[i]->descriptor()->instantiate
                        (myType->pluginTypes[i]->descriptor(), sampleRate));
    }
};


/******************************************************************************
 *                            PluginChainType                                 *
 ******************************************************************************/

PluginChainType::PluginChainType (const char* spec)
{
    parseXML (spec);
    classifyPorts ();
    // Create the virtual Descriptor
    createDescriptorMiscData ();
    setDescriptorFunctionPointers ();
    createDescriptorPortsData ();
};

PluginChainType::~PluginChainType ()
{
    free (const_cast<char*>(theDescriptor.Label));
    free (const_cast<char*>(theDescriptor.Name));
    for (unsigned long i = 0; i < theDescriptor.PortCount; ++i) {
        free (const_cast<char**>(theDescriptor.PortNames)[i]);
    }
    delete[] theDescriptor.PortDescriptors;
    delete[] theDescriptor.PortNames;
    delete[] theDescriptor.PortRangeHints;
    for (unsigned long i = 0; i < pluginTypes.size(); ++i) {
        delete pluginTypes[i];
    }
};

void
PluginChainType::classifyPorts ()
{
    vector<pair<unsigned long, unsigned long> > ins, outs, extraIns,
                                                extraOuts, controls;
    for (unsigned long i = 0; i < pluginTypes.size(); ++i) {
        unsigned long numIns, numOuts;
        numIns = numOuts = 0;
        const LADSPA_Descriptor* descriptor (pluginTypes[i]->descriptor());
        if (LADSPA_IS_INPLACE_BROKEN (descriptor->Properties)) {
            // Since we are traversing the plugins, let's collect
            // the number of InPlaceBroken ones.
            ++numInPlaceBroken;
        }

        for (unsigned long j = 0; j < descriptor->PortCount; ++j) {
            if (LADSPA_IS_PORT_AUDIO (descriptor->PortDescriptors[j])) {
                if (LADSPA_IS_PORT_INPUT (descriptor->PortDescriptors[j])) {
                    ++numIns;
                    if (numIns > numChannels) {
                        extraIns.push_back (make_pair (i, j));
                    } else {
                        ins.push_back (make_pair (i, j));
                    }
                } else {
                    ++numOuts;
                    if (numOuts > numChannels) {
                        extraOuts.push_back (make_pair (i, j));
                    } else {
                        outs.push_back (make_pair (i, j));
                    }
                }
            } else {
                controls.push_back (make_pair(i, j));
            }
        }
    }
    channelPortsSequence.resize (numChannels);
    for (unsigned long i = 0; i < pluginTypes.size(); ++i) {
        for (unsigned long j = 0; j < numChannels; ++j) {
            channelPortsSequence[j].push_back
                        (make_pair (ins[i*numChannels+j].second,
                        outs[i*numChannels+j].second));
        }
    }
    // Insert the inputs of the first plugin and the outputs of the
    // last plugins as the inputs and outputs of the chain.
    for (unsigned long i = 0; i < numChannels; ++i) {
        virtualPorts.push_back (make_pair (0, ins[i].second));
    }
    for (unsigned long i = 0; i < numChannels; ++i) {
        virtualPorts.push_back (make_pair
                                    (pluginTypes.size() -1, outs[i].second));
    }
    virtualPorts.insert (virtualPorts.end(), controls.begin(), controls.end());
    virtualPorts.insert (virtualPorts.end(), extraIns.begin(), extraIns.end());
    virtualPorts.insert (virtualPorts.end(), extraOuts.begin(), extraOuts.end());
};

void
PluginChainType::createDescriptorMiscData ()
{
    // FIXME!!!!! We need a way to deal with UniqueID cleanly!
    theDescriptor.UniqueID = 999;

    if (pluginTypes.size() == 1) {
        theDescriptor.Label = strdup ((*(pluginTypes.begin()))->descriptor()->Label);
        theDescriptor.Name = strdup ((*(pluginTypes.begin()))->descriptor()->Name);
    } else {
        ostringstream s;
        s << "Chained";
        for (vector<MultiChannelPluginType*>::iterator it =
                    pluginTypes.begin(); it != pluginTypes.end(); ++it) {
            s << "_";
            s << (*it)->descriptor()->Label;
        }
        theDescriptor.Label = strdup (s.str().c_str());
        theDescriptor.Name = strdup (s.str().c_str());
    }

    bool realtime = false;
    bool RTCapable = true;
    for (vector<MultiChannelPluginType*>::iterator it =
                    pluginTypes.begin(); it != pluginTypes.end(); ++it) {
        LADSPA_Properties pluginProperties ((*it)->descriptor()->Properties);
        if (LADSPA_IS_REALTIME (pluginProperties)) {
            realtime = true;
        }
        if (! LADSPA_IS_HARD_RT_CAPABLE (pluginProperties)) {
           RTCapable = false;
        }
    }
    bool inPlaceBroken = false;
    // Since the plugins may have extra ports, it's better
    // to be InPlaceBroken if the first or last plugin is
    // InPlaceBroken. Also, if the first/last plugin is
    // InPlaceBroken, it may be difficult to do the processing
    // correctly with only a single intermediate buffer if
    // the input and the output are the same.
    if (LADSPA_IS_INPLACE_BROKEN
                        (pluginTypes.front()->descriptor()->Properties) ||
                        LADSPA_IS_INPLACE_BROKEN
                        (pluginTypes.back()->descriptor()->Properties)) {
        inPlaceBroken = true;
    }

    if (RTCapable) {
        theDescriptor.Properties = LADSPA_PROPERTY_HARD_RT_CAPABLE;
    } else {
        theDescriptor.Properties = 0;
    }
    if (inPlaceBroken) {
        theDescriptor.Properties |= LADSPA_PROPERTY_INPLACE_BROKEN;
    }
    if (realtime) {
        theDescriptor.Properties |= LADSPA_PROPERTY_REALTIME;
    }

    theDescriptor.Maker = "Nelson Lago";
    theDescriptor.Copyright = "GPL";

    // When we instantiate the plugin, we need to access this instance
    // of the MultiChannelPluginType to know what the plugin to be
    // instantiated looks like.
    theDescriptor.ImplementationData = this;
};

void
PluginChainType::setDescriptorFunctionPointers ()
{
    theDescriptor.instantiate = ::instantiate<PluginChain>;
    theDescriptor.connect_port = ::connect_port<PluginChain>;
    theDescriptor.activate = ::activate<PluginChain>;
    theDescriptor.run = ::run<PluginChain>;
    // FIXME We should deal with run_adding.
    theDescriptor.run_adding = 0;
    theDescriptor.set_run_adding_gain = 0;
    theDescriptor.deactivate = ::deactivate<PluginChain>;
    theDescriptor.cleanup = ::cleanup<PluginChain>;
};

void
PluginChainType::createDescriptorPortsData ()
{
    // Now create the virtual ports
    unsigned long numPorts = virtualPorts.size();
    theDescriptor.PortCount = numPorts;
    theDescriptor.PortDescriptors =
                new LADSPA_PortDescriptor[numPorts];
    theDescriptor.PortNames = new const char* [numPorts];
    theDescriptor.PortRangeHints =
                new LADSPA_PortRangeHint[numPorts];

    // Define each port's info: name, hints etc. Just copy the
    // data from the real plugins.
    for (unsigned long virtualPort = 0; virtualPort < virtualPorts.size();
                                                        ++virtualPort) {
        unsigned long pluginPort, plugin;
        plugin = virtualPorts[virtualPort].first;
        pluginPort = virtualPorts[virtualPort].second;
        // Just copy the original data
        const_cast<LADSPA_PortDescriptor*>
            (theDescriptor.PortDescriptors)[virtualPort] =
                        pluginTypes[plugin]->descriptor()->PortDescriptors[pluginPort];
        const_cast<LADSPA_PortRangeHint*>
                (theDescriptor.PortRangeHints)[virtualPort] =
                        pluginTypes[plugin]->descriptor()->PortRangeHints[pluginPort];
        const_cast<char**> (theDescriptor.PortNames)[virtualPort] =
            strdup (pluginTypes[plugin]->descriptor()->PortNames[pluginPort]);
    }
};

// FIXME!!!
// AAArgh!! All these "reinterpret_cast"s suck so badly I have no
// words to describe. Still, they work ;-)

void
PluginChainType::parseXML (const char* spec)
{
    xmlDocPtr doc;
      doc = xmlParseMemory (spec, strlen (spec));
    if (doc == 0) {
        // FIXME
        cerr << "PluginChainType: unable to parse XML" << endl;
        exit (1);
    }
    if (xmlStrcmp (reinterpret_cast<xmlDtdPtr>(doc)->children->name,
                reinterpret_cast<const xmlChar*>("jackrack")) != 0) {
        // FIXME
        cerr << "PluginChainType: invalid XML" << endl;
        exit (1);
    }
    xmlNodePtr jackrack;
    for (jackrack = doc->children; jackrack != 0; jackrack = jackrack->next) {
        if (xmlStrcmp (jackrack->name,
                        reinterpret_cast<const xmlChar*>("jackrack")) == 0) {
            xmlNodePtr node;
            xmlChar *content;
            for (node = jackrack->children; node != 0; node = node->next) {
                if (xmlStrcmp (node->name,
                        reinterpret_cast<const xmlChar*>("channels")) == 0) {
                    content = xmlNodeGetContent (node);
                    numChannels = strtoul
                            (reinterpret_cast<const char*>(content), 0, 10);
                    xmlFree (content);
                } else if (xmlStrcmp (node->name,
                        reinterpret_cast<const xmlChar*>("samplerate")) == 0) {
                    // We ignore the sampleRate here; it's only useful
                    // when we are actually going to instantiate a plugin.
                    continue;
                } else if (xmlStrcmp (node->name,
                        reinterpret_cast<const xmlChar*>("plugin")) == 0) {
                    parsePlugin (node);
                }
            }
        }
    }
    xmlFreeDoc (doc);
};

void
PluginChainType::parsePlugin (xmlNodePtr plugin)
{
    xmlNodePtr node;
    xmlChar *content;
    unsigned long pluginID = 0;
    bool enabled = false;
    PortOptions wetnessOptions;
    wetnessOptions.enabled = false;
    wetnessOptions.locked = false;
    vector<PortOptions> controlOptions;
    bool allLocked = false;

    for (node = plugin->children; node != 0; node = node->next) {
        if (xmlStrcmp (node->name,
                    reinterpret_cast<const xmlChar*>("id")) == 0) {
            content = xmlNodeGetContent (node);
            pluginID = strtoul (reinterpret_cast<const char*>(content), 0, 10);
            xmlFree (content);
        } else if (xmlStrcmp (node->name,
                        reinterpret_cast<const xmlChar*>("enabled")) == 0) {
            content = xmlNodeGetContent (node);
            if (xmlStrcmp (content,
                        reinterpret_cast<const xmlChar*>("true")) == 0) {
                enabled = true;
            }
            xmlFree (content);
        } else if (xmlStrcmp (node->name,
                    reinterpret_cast<const xmlChar*>("wet_dry_enabled")) == 0) {
            content = xmlNodeGetContent (node);
            if (xmlStrcmp (content,
                        reinterpret_cast<const xmlChar*>("true")) == 0) {
                wetnessOptions.enabled = true;
            }
            xmlFree (content);
        } else if (xmlStrcmp (node->name,
                    reinterpret_cast<const xmlChar*>("wet_dry_locked")) == 0) {
            content = xmlNodeGetContent (node);
            if (xmlStrcmp (content,
                        reinterpret_cast<const xmlChar*>("true")) == 0) {
                wetnessOptions.locked = true;
            }
            xmlFree (content);
        } else if (xmlStrcmp (node->name,
                    reinterpret_cast<const xmlChar*>("wet_dry_values")) == 0) {
            xmlNodePtr subnode;
            for (subnode = node->children; subnode != 0;
                                                subnode = subnode->next) {
                if (xmlStrcmp (subnode->name,
                            reinterpret_cast<const xmlChar*>("value")) == 0) {
                    content = xmlNodeGetContent (subnode);
                    wetnessOptions.values.push_back (strtod (
                                reinterpret_cast<const char*>(content), 0));
                    xmlFree (content);
                }
            }
        } else if (xmlStrcmp (node->name,
                        reinterpret_cast<const xmlChar*>("lockall")) == 0) {
            content = xmlNodeGetContent (node);
            if (xmlStrcmp (content,
                        reinterpret_cast<const xmlChar*>("true")) == 0) {
                allLocked = true;
            }
            xmlFree (content);
        } else if (xmlStrcmp (node->name,
                    reinterpret_cast<const xmlChar*>("controlrow")) == 0) {
            xmlNodePtr subnode;
            controlOptions.push_back (PortOptions());
            controlOptions.back().enabled = true;
            controlOptions.back().locked = false;
            for (subnode = node->children; subnode != 0;
                                                subnode = subnode->next) {
                if (xmlStrcmp (subnode->name,
                            reinterpret_cast<const xmlChar*>("lock")) == 0) {
                    content = xmlNodeGetContent (node);
                    if (xmlStrcmp (content,
                            reinterpret_cast<const xmlChar*>("true")) == 0) {
                        controlOptions.back().locked = true;
                    }
                    xmlFree (content);
                } else if (xmlStrcmp (subnode->name,
                            reinterpret_cast<const xmlChar*>("value")) == 0) {
                    content = xmlNodeGetContent (node);
                    controlOptions.back().values.push_back (strtod(
                                reinterpret_cast<const char*>(content), 0));
                    xmlFree (content);
                }
            }
        }
    }
    if (allLocked) {
        vector<PortOptions>::iterator it;
        for (it = controlOptions.begin(); it != controlOptions.end(); ++it) {
            it->locked = true;
        }
    };
    // All set! Let's create the MultiChannelPluginType
    if (enabled) {
        pluginTypes.push_back (new MultiChannelPluginType
                (pluginID, numChannels, controlOptions, wetnessOptions));
    }
};

} // namespace
