#ifndef DLADSPA_PLUGIN_CHAIN_H
#define DLADSPA_PLUGIN_CHAIN_H

/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * PluginChain wraps a sequence of LADSPA plugins and presents it to the
 * client as a single LADSPA plugin with all the control ports. for instance,
 * we may use it to create a "dark distortion" plugin simply by using a
 * distortion plugin followed by a low-pass filter plugin. The dark distortion
 * plugin would have the control ports to control the amount of distortion
 * and the cutoff frequency for the filter, controlling the "darkness".
 * 
 * PluginChainType receives the data on the kind of PluginChain we want to
 * create on the constructor as a XML description (in the format of the
 * files saved by the jack-rack application) and creates the corresponding
 * LADSPA_Descriptor and some auxiliary data. On the created LADSPA_Descriptor,
 * PluginChainType puts a reference to itself in ImplementationData;
 * this is used when the client calls instantiate on the plugin (it's
 * the only way PluginChain "knows" how it should be created). This means
 * that as long as the LADSPA_Descriptor is in use, PluginChainType must
 * remain a valid object.
 ******************************************************************************/ 


#include <ladspa.h>

#include <dladspa/multi_channel_plugin.h>

// FIXME! It should be possible not to include this here, but only
// in plugin_chain.cc instead.
#include <libxml/tree.h>


namespace DLADSPA {

class PluginChainType {
friend class PluginChain;
public:
    PluginChainType (const char*);
    ~PluginChainType ();
    inline const LADSPA_Descriptor* descriptor () const;
private:
    LADSPA_Descriptor theDescriptor;
    std::vector<MultiChannelPluginType*> pluginTypes;
    unsigned long numChannels, numInPlaceBroken;
    // This maps the virtual port numbers to the corresponding
    // plugin and its port. The vector index is the virtual port
    // number and the pair parameters are the plugin number and
    // the plugin's port number.
    std::vector<std::pair<unsigned long, unsigned long> > virtualPorts;
    // A channel to be processed consists of a collection of
    // specific input and output ports for each of the plugins
    // in the chain. This vector stores, for each channel, such
    // a collection, in the form of another vector. In this
    // secondary vector, the index corresponds to the plugin
    // number, the first pair parameter to the plugin's input
    // and the second pair parameter to the corresponding
    // plugin's output.
    std::vector<std::vector<std::pair<unsigned long, unsigned long> > >
                                                        channelPortsSequence;

    void parseXML(const char*);
    void parsePlugin (xmlNodePtr);
    void classifyPorts ();
    void createDescriptorMiscData ();
    void setDescriptorFunctionPointers ();
    void createDescriptorPortsData();
};

class PluginChain {
public:
    PluginChain (const LADSPA_Descriptor* desc, unsigned long sampleRate);
    ~PluginChain();
    inline void activate ();
    inline void deactivate ();
    void connect_port (unsigned long port, LADSPA_Data* ptr);
    inline void run (unsigned long sampleCount);
    inline void run_adding (unsigned long sampleCount);
    inline void set_run_adding_gain (LADSPA_Data gain);
    inline void cleanup ();
private:
    PluginChainType* myType;
    std::vector<LADSPA_Handle> plugins;
    std::vector<LADSPA_Data*> inputAddresses, outputAddresses, buffers;
    void connectChannel (unsigned long channel);
};


/******************************************************************************
 *                               Inlines
 ******************************************************************************/

inline const LADSPA_Descriptor*
PluginChainType::descriptor () const
{
    return &theDescriptor;
};

inline void
PluginChain::activate ()
{
    for (unsigned long i = 0; i < plugins.size(); ++i) {
        if (myType->pluginTypes[i]->descriptor()->activate != 0) {
            myType->pluginTypes[i]->descriptor()->activate(plugins[i]);
        }
    }
};
inline void
PluginChain::deactivate ()
{
    for (unsigned long i = 0; i < plugins.size(); ++i) {
        if (myType->pluginTypes[i]->descriptor()->deactivate != 0) {
            myType->pluginTypes[i]->descriptor()->deactivate(plugins[i]);
        }
    }
};

inline void
PluginChain::run (unsigned long sampleCount)
{
    for (unsigned long i = 0; i < plugins.size(); ++i) {
        myType->pluginTypes[i]->descriptor()->run(plugins[i], sampleCount);
    }
};

inline void
PluginChain::run_adding (unsigned long sampleCount)
{
    for (unsigned long i = 0; i < plugins.size(); ++i) {
        if (myType->pluginTypes[i]->descriptor()->run_adding != 0) {
            myType->pluginTypes[i]->descriptor()->run_adding
                                            (plugins[i], sampleCount);
        }
    }
};

inline void
PluginChain::set_run_adding_gain (LADSPA_Data gain)
{
    for (unsigned long i = 0; i < plugins.size(); ++i) {
        if (myType->pluginTypes[i]->descriptor()->set_run_adding_gain != 0) {
            myType->pluginTypes[i]->descriptor()->set_run_adding_gain
                                            (plugins[i], gain);
        }
    }
};

inline void
PluginChain::cleanup ()
{
    // If the plugin doesn't offer this function, there's a memory leak.
    for (unsigned long i = 0; i < plugins.size(); ++i) {
        if (myType->pluginTypes[i]->descriptor()->cleanup != 0) {
            myType->pluginTypes[i]->descriptor()->cleanup(plugins[i]);
        }
    }
};

} // namespace
#endif // ifndef DLADSPA_PLUGIN_CHAIN_H
