/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


#include <dladspa/constants.h>
#include <dladspa/proxy.h>
#include <dlowlat/masterslave/masterslave.h>

#include <dlowlat/transport/ip/tcp/sockets.h>

#include <sstream>
#include <fstream>
#include <stdlib.h>

using namespace DLowLat;

namespace DLADSPA {

Proxy::Proxy (const LADSPA_Descriptor* desc,
                        unsigned long sampleRate) :
                        descriptor (desc)
{
    this->sampleRate = sampleRate;

    setupConnection ();

    constantsSize = 1;
    sampleCount = 0;
    runAddingGain = 1;
    functionToCall = FUNCTION_RUN;
    
    theBlocks.push_back (processor->newDataBlock());
    theBlocks.back()->setBlock (&functionToCall, constantsSize, OUT, false);

    theBlocks.push_back (processor->newDataBlock());
    theBlocks.back()->setBlock (&sampleCount, constantsSize, OUT, false);

    // Create the placeholder for the LADSPA ports
    setupPorts();
};

void
Proxy::setupConnection()
{
    // Find the IP address/port of the manager program
    char* managerEnv = getenv (MANAGER_ENV_NAME);
    if (managerEnv == 0) {
        // Default
        managerEnv = MANAGER_DEFAULT_ADDRESS;
    }
    managerAddress.host (managerEnv);

    // connect to the manager and request the creation of the plugin
    TcpClient manager (managerAddress);
    IPAddress myAddress;
    UdpRead reader (myAddress, false);
    manager << "CREATE";
    manager << myAddress;
    manager << theXMLDescriptor();
    manager << sampleRate;
    // get answer
    string s;
    manager >> s;
    if (! (s == "CREATED")) {
        cerr << "The remote plugin was not created!" << endl;
        // FIXME
        exit (1);
    }
    processorHostname = "";
    manager >> processorHostname;
    int windowSize;
    manager >> windowSize;
    IPAddress peerAddress;
    manager >> peerAddress;
    UdpWrite writer (peerAddress);
    // Create the local processor object, which sends data to
    // be remotely processed
    processor = new Master (writer, reader, LITTLEENDIAN, windowSize);
};

void
Proxy::setupPorts ()
{
    for (unsigned int i = 0; i < descriptor->PortCount; ++i) {
        theBlocks.push_back (processor->newDataBlock());
        Direction dir = NONE;
        // FIXME! This naming is WEIRD! LADSPA Input ports are ports that
        // we need to send out to our peer
        isIn (dir, LADSPA_IS_PORT_OUTPUT(descriptor->PortDescriptors[i]));
        isOut (dir, LADSPA_IS_PORT_INPUT(descriptor->PortDescriptors[i]));
        if (LADSPA_IS_PORT_CONTROL (descriptor->PortDescriptors[i])) {
            theBlocks.back()->setBlock (static_cast<LADSPA_Data*>(0), constantsSize, dir, false);
        } else {
            theBlocks.back()->setBlock (static_cast<LADSPA_Data*>(0), sampleCount, dir, true);
        }
    }
};

Proxy::~Proxy ()
{
    // conect to the manager and request the destruction of the plugin
    TcpClient manager (managerAddress);
    manager << "DELETE" << processorHostname;
    manager << processor->peerAddress();
    // Delete the local processor, which sends data to be remotely processed
    delete processor;
};

void
Proxy::connect_port (unsigned long port, LADSPA_Data* ptr)
{
    theBlocks[port + NUM_AUTOMATIC_ELEMENTS]->blockAddress (ptr);
};

void
Proxy::processData (unsigned long sampleCount)
{
    // We can't really cope gracefully with variable sampleCounts
    // in general, but since we only know *any* value of
    // sampleCount on the first call to this function, this
    // code is unavoidable.
    this->sampleCount = sampleCount;
    processor->process ();
};

void
Proxy::activate ()
{
    // We don't need to activate the remote plugin here; it is
    // always active.
    sampleCount = 0;

    // If we were previously active then deactivated, we need to
    // dump old data received from the network if it exists.
    processor->reset();
};

void
Proxy::deactivate ()
{
    // We always keep the remote plugin activated; this function,
    // therefore, only resets the remote plugin, by calling
    // deactivate() and activate() on it.

    // Let's avoid reading data from the LADSPA ports, since the
    // buffers may have been freed() by the host or something.
    sampleCount = 0;

    functionToCall = FUNCTION_RESET;
    processor->process();
};

string
Proxy::theXMLDescriptor()
{
    // Find the directory where the xml plugin descriptions are
    char* xmlEnv = getenv (XML_PLUGIN_PATH_ENV_NAME);
    if (xmlEnv == 0) {
        // Default
        xmlEnv = DEFAULT_XML_PLUGIN_PATH;
    }
    string theFilename (xmlEnv);
    theFilename += '/';
    ostringstream s;
    s << descriptor->UniqueID;
    theFilename += s.str();
    theFilename += ".rack";
    ifstream xmlfile (theFilename.c_str());
    // I guess there's a better way to do this, but who cares?
    s.str ("");
    s << xmlfile.rdbuf();
    return s.str();
};

} // namespace
