#ifndef DLADSPA_PROXY_H
#define DLADSPA_PROXY_H

/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * This is a proxy class for LADSPA plugins. For each plugin we want to
 * proxy we need to create a LADSPA_Descriptor that lists the characteristics
 * of the (remote) plugin; so, this LADSPA_Descriptor is almost a copy of
 * the real plugin's descriptor. However, the callback functions point to
 * the generic functions defined in function_pointers.h which, in turn, call
 * methods in this class.
 *
 * a single proxy class can work for any plugin: that's because we just need
 * to grab the data and pass it along, so the proxy class doesn't need any
 * kind of specialization according to the plugin type.
 *
 * This class runs inside a "host" application that knows nothing about
 * the mechanics of communication used here; therefore, we can't use
 * command line parameters to inform this class where is the connection
 * manager; we use environment variables instead, defined in constants.h.
 ******************************************************************************/

// Files that #include this one may also need to include:
// <dlowlat/masterslave/masterslave.h>

#include <ladspa_sdk/ladspa.h>
#include <dladspa/constants.h>

#include <vector>
#include <string>

using namespace std;

#include <dlowlat/transport/ip/ipaddress.h>
#include <dlowlat/masterslave/datablocks.h>

// Some forward declarations needed by this file:
namespace DLowLat {
    class Master;
}

namespace DLADSPA {

/******************************************************************************
 *                         The proxy class
 ******************************************************************************/

class Proxy {
public:
    Proxy (const LADSPA_Descriptor* descriptor,
                unsigned long sampleRate);
    void connect_port (unsigned long port, LADSPA_Data* ptr);
    void activate ();
    inline void run (unsigned long sampleCount);
    inline void run_adding (unsigned long sampleCount);
    inline void set_run_adding_gain (LADSPA_Data gain);
    void deactivate (); // Temporarily disable plugin
    ~Proxy();
private:
    size_t constantsSize;
    void setupConnection();
    void setupPorts();
    void processData (unsigned long sampleCount);
    DLowLat::Master* processor;
    size_t sampleCount;
    unsigned long sampleRate;
    LADSPA_Data runAddingGain;
    // This is passed to the remote end together with the data
    // from the LADSPA ports.
    unsigned int functionToCall;
    vector<DLowLat::DataBlock> theBlocks;
    const LADSPA_Descriptor* descriptor;
    string processorHostname;
    DLowLat::IPAddress managerAddress;
    // This reads the file with the XML that corresponds to the descriptor
    // we are dealing with
    string theXMLDescriptor();
};


/******************************************************************************
 *                                  inlines
 ******************************************************************************/

inline void
Proxy::run (unsigned long sampleCount)
{
    functionToCall = FUNCTION_RUN;
    processData (sampleCount);
};

inline void
Proxy::run_adding (unsigned long sampleCount)
{
    // FIXME Since LADSPA allows us not to provide this function, that's
    // what we should do; still, it could be possible to implement it.
    processData (sampleCount);
};

inline void
Proxy::set_run_adding_gain (LADSPA_Data gain)
{
    runAddingGain = gain;
};

} // namespace

#endif
