/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/

/******************************************************************************
 * This file is to be compiled as a dll that behaves like any other LADSPA dll:
 * when dlopen()'ed, the function ladspa_descriptor() may be used to get
 * information about available plugins, in the form of LADSPA_Descriptors. In
 * our case, though, these descriptors actually describe proxy plugins,
 * implemented by the DLADSPA::Proxy class. To do this, this file instantiates a
 * Loader object when loaded (the "init" object), which loads plugin
 * descriptions from files and creates the corresponding LADSPA_Descriptors.
 ******************************************************************************/


#include <sys/types.h>
#include <dirent.h>

#include <fstream>
#include <sstream>
#include <vector>

#include <dladspa/constants.h>
#include <dladspa/plugin_chain.h>
#include <dladspa/function_pointers.h>
#include <dladspa/proxy.h>

#include <iostream>

extern "C" {
#include <ladspa_sdk/ladspa.h>
}

using namespace DLADSPA;

// Let's hide the stuff that is not part of the LADSPA interface (that is,
// the ladspa_descriptor() function) in an unnamed namespace.

namespace {

/******************************************************************************
 *                         The Loader Class
 ******************************************************************************/

class Loader {
public:
    Loader ();
    ~Loader ();
    inline const LADSPA_Descriptor* operator[] (unsigned int);
private:
    void readPlugin (const string& filename, unsigned long pluginNumber);
    vector<PluginChainType*> pluginTypes;
};

Loader::~Loader()
{
    for (unsigned i = 0; i < pluginTypes.size(); ++i) {
        delete pluginTypes[i];
    }
};

Loader::Loader ()
{
    // Find the directory where the xml plugin descriptions are
    char* xmlPath = getenv (XML_PLUGIN_PATH_ENV_NAME);
    if (xmlPath == 0) {
        // Default
        xmlPath = DEFAULT_XML_PLUGIN_PATH;
    }
    // Read all files in directory and create the corresponding
    // pluginTypes.
    DIR* pluginDir = opendir (xmlPath);

    if (pluginDir == 0) {
      cerr << "remote_plugins: unable to open directory: " << xmlPath << endl;
      exit (1);
    }
    struct dirent* pluginfile;
    while ((pluginfile = readdir (pluginDir)) != 0) {
        string filename (pluginfile->d_name);
        string::size_type ext;
        if ((ext = filename.find (".rack")) == string::npos) {
            // This is not a jack-rack file
            continue;
        }
        istringstream s (filename.substr (0, ext));
        unsigned long pluginNumber;
        s >> pluginNumber;

        filename = "/" + filename;
        filename = xmlPath + filename;
        readPlugin (filename, pluginNumber);
    }
    closedir (pluginDir);
};

void
Loader::readPlugin (const string& filename, unsigned long pluginNumber)
{
    ifstream xmlfile (filename.c_str());
    ostringstream s;
    s << xmlfile.rdbuf();
    pluginTypes.push_back (new PluginChainType (s.str().c_str()));

    // FIXME!! This is an ugly way to do this, but is simple.
    // We should follow the design of the MultiChannelPlugin and
    // PluginChain classes, defining a "ProxyType" class that
    // creates the proxy descriptors.
    LADSPA_Descriptor* ptr =
    const_cast<LADSPA_Descriptor*>(pluginTypes.back()->descriptor());
    // Since, at each iteration, the proxy plugin outputs data from a
    // previous iteration, it needs to operate in realtime on most situations.
    ptr->Properties |= LADSPA_PROPERTY_REALTIME;
    ptr->UniqueID = pluginNumber;
    ptr->instantiate = ::instantiate<Proxy>;
    ptr->connect_port = ::connect_port<Proxy>;
    ptr->activate = ::activate<Proxy>;
    ptr->run = ::run<Proxy>;
    // FIXME We should deal with run_adding.
    ptr->run_adding = 0;
    ptr->set_run_adding_gain = 0;
    ptr->deactivate = ::deactivate<Proxy>;
    ptr->cleanup = ::cleanup<Proxy>;
};

inline const LADSPA_Descriptor*
Loader::operator[] (unsigned int i)
{
    if (i >= pluginTypes.size()) {
        return 0;
    } else {
        return pluginTypes[i]->descriptor();
    }
};


/******************************************************************************
 *            load the plugin files and create the descriptors
 ******************************************************************************/

// All the work is done by the Loader constructor; isn't that pretty? ;-)
Loader init;

} // Close the unnamed namespace


// Clients will use this function to query which are the available plugins
extern "C"
{
    const LADSPA_Descriptor* ladspa_descriptor (unsigned long i)
    {
        return init [i];
    }
} // extern "C"
