/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * This file is to be compiled as a dll that behaves like any other LADSPA dll:
 * when dlopen()'ed, the function ladspa_descriptor() may be used to get
 * information about available plugins, in the form of LADSPA_Descriptors. In
 * our case, there is a single plugin: a test plugin that copies its inputs to
 * its outputs and then busy-waits for some time. While it operates, it gets
 * some statistics from linux's /proc pseudo-filesystem.
 ******************************************************************************/


using namespace std;

#include <ladspa.h>

#include <miscutils/cronometer.h>
#include <miscutils/cpu_stats.h>

#include <iostream>
#include <sstream>

#include <string.h>

namespace {

enum {
    MAX_TIME = 0,
    MANUAL_MODE = 1,
    DATA_INPUT = 2,
    DATA_OUTPUT = 3,
    SINGLE_TEST_TIME = 10000000 // 10 seconds
};

typedef struct {
    unsigned long sleepTime;
    int userTimePercent;
    int systemTimePercent;
    int idleTimePercent;
    double collisionsPercent;
} StatData;

typedef struct {
    Cronometer *timer, *waiter;
    CPUStats *statsGrabber;
    unsigned long currentSleepTime;
    bool hasntRunYet;
    int statIndex;
    unsigned long blockSize;
    LADSPA_Data* maxTime;
    LADSPA_Data* manualMode;
    LADSPA_Data* dataInput;
    LADSPA_Data* dataOutput;
    StatData* statistics;
} InstanceData;
    
LADSPA_Handle
instantiate (const LADSPA_Descriptor* descriptor, unsigned long sampleRate)
{
    InstanceData* instance = new InstanceData ();
    instance->timer = new Cronometer();
    instance->waiter = new Cronometer();
    instance->statsGrabber = new CPUStats();
    instance->statistics = new StatData[100];
    return instance;
};

void
activate (LADSPA_Handle handle)
{
    InstanceData* instance = static_cast<InstanceData*>(handle);
    instance->statIndex = 0;
    instance->currentSleepTime = 1; // 100 useconds
    instance->hasntRunYet = true;
    for (int i = 0; i < 100; ++i) {
        instance->statistics[i].sleepTime = 0;
    }
};

void
recordStats (InstanceData& instance)
{
    instance.statistics[instance.statIndex].sleepTime =
                        instance.currentSleepTime;

    instance.statistics[instance.statIndex].userTimePercent =
                        instance.statsGrabber->userTimePercent();

    instance.statistics[instance.statIndex].systemTimePercent =
                        instance.statsGrabber->systemTimePercent();

    instance.statistics[instance.statIndex].idleTimePercent =
                        instance.statsGrabber->idleTimePercent();

    instance.statistics[instance.statIndex].collisionsPercent =
                        instance.statsGrabber->collisionsPercent();

    if (instance.statIndex < 99) {
        instance.statIndex++;
    }
}

void
connect_port (LADSPA_Handle handle, unsigned long port, LADSPA_Data* location)
{
    InstanceData* instance = static_cast<InstanceData*>(handle);
    switch (port) {
        case MAX_TIME:
            instance->maxTime = location;
            break;
        case MANUAL_MODE:
            instance->manualMode = location;
            break;
        case DATA_INPUT:
            instance->dataInput = location;
            break;
        case DATA_OUTPUT:
            instance->dataOutput = location;
            break;
    }
}

void
run (LADSPA_Handle handle, unsigned long count)
{
    InstanceData* instance = static_cast<InstanceData*>(handle);

    instance->waiter->start();

    if (instance->hasntRunYet) {
        instance->timer->start();
        instance->statsGrabber->start();
        instance->hasntRunYet = false;
        instance->blockSize = count;
    }

    for (unsigned int i = 0; i < count; ++i) {
        (instance->dataOutput)[i] = (instance->dataInput)[i];
    }

    if (*(instance->manualMode) > 0.1) {
        instance->currentSleepTime = static_cast<long>(*(instance->maxTime));
    }

    if (instance->timer->elapsed() > SINGLE_TEST_TIME) {
        instance->statsGrabber->stop();
        recordStats (*instance);

        if (instance->currentSleepTime < static_cast<unsigned long>(*(instance->maxTime))) {
            instance->currentSleepTime++; // add 100usecs to the test time
        }
        instance->statsGrabber->start();
        instance->timer->start();
    }

    // We don't want to wait a negative time, specially since these are
    // unsigneds; so, if we're only 20usecs before the timeout, we just
    // exit.
    if (instance->currentSleepTime * 100 > instance->waiter->elapsed() + 20) {
        instance->waiter->busyWait (100 * instance->currentSleepTime - instance->waiter->elapsed());
    }
}

void writeStats (StatData* stats, unsigned long count)
{
    ostringstream s;
    s << "/tmp/bufsize_" << count << ".log";
    ofstream f (s.str().c_str());
    f << "sleepTime | utime | stime | itime | coll" << endl;
    for (int i = 0; i < 100; ++i)
    {
        if (stats[i].sleepTime > 0) {
            f << stats[i].sleepTime << " " << stats[i].userTimePercent << " ";
            f << stats[i].systemTimePercent << " " << stats[i].idleTimePercent;
            f << " " << stats[i].collisionsPercent << endl;
        }
    }
    f << "----------------------------" << endl;
};

void
cleanup (LADSPA_Handle handle)
{
    InstanceData* instance = static_cast<InstanceData*>(handle);
    writeStats (instance->statistics, instance->blockSize);
    delete instance->waiter;
    delete instance->timer;
    delete instance->statsGrabber;
    delete[] instance->statistics;
    delete instance;
}

void deactivate (LADSPA_Handle handle)
{
    InstanceData* instance = static_cast<InstanceData*>(handle);
    writeStats (instance->statistics, instance->blockSize);
}

class Loader
{
public:
    Loader ();
    const LADSPA_Descriptor* theDescriptor () { return &descriptor; };
private:
    LADSPA_Descriptor descriptor;
};

// FIXME: We need a destructor to clean up these C-style strings; maybe we
// could not use strcpy, but just point to the string directly? Why was this
// done this way?
Loader::Loader() {
    descriptor.UniqueID = 1000;
    char* s;
    s = new char[strlen("waste_time")+1];
    strcpy (s, "waste_time");
    descriptor.Label = s;
    descriptor.Properties = LADSPA_PROPERTY_HARD_RT_CAPABLE;
    s = new char[strlen("Waste time test plugin")+1];
    strcpy (s, "Waste time test plugin");
    descriptor.Name = s;
    s = new char[strlen("Nelson Lago")+1];
    strcpy (s, "Nelson Lago");
    descriptor.Maker = s;
    s = new char[strlen("GPL")+1];
    strcpy (s, "GPL");
    descriptor.Copyright = s;
    descriptor.PortCount = 4;

    LADSPA_PortDescriptor* portDescriptors = new LADSPA_PortDescriptor[4];
    portDescriptors[MAX_TIME] = LADSPA_PORT_CONTROL | LADSPA_PORT_INPUT;
    portDescriptors[MANUAL_MODE] = LADSPA_PORT_CONTROL | LADSPA_PORT_INPUT;
    portDescriptors[DATA_INPUT] = LADSPA_PORT_AUDIO | LADSPA_PORT_INPUT;
    portDescriptors[DATA_OUTPUT] = LADSPA_PORT_AUDIO | LADSPA_PORT_OUTPUT;
    descriptor.PortDescriptors = portDescriptors;

    char** portNames = new char*[4];
    portNames[MAX_TIME] = new char[strlen("Maximum wait time (hundreds of microseconds)")+1];
    strcpy (portNames[MAX_TIME], "Maximum wait time (hundreds of microseconds)");
    portNames[MANUAL_MODE] = new char[strlen("Manual mode")+1];
    strcpy (portNames[MANUAL_MODE], "Manual mode");
    portNames[DATA_INPUT] = new char[strlen("Input")+1];
    strcpy (portNames[DATA_INPUT], "Input");
    portNames[DATA_OUTPUT] = new char[strlen("Output")+1];
    strcpy (portNames[DATA_OUTPUT], "Output");
    descriptor.PortNames = portNames;

    LADSPA_PortRangeHint* portRangeHints = new LADSPA_PortRangeHint[4];
    portRangeHints[MAX_TIME].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW |
                            LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_INTEGER;
    portRangeHints[MAX_TIME].LowerBound = 0;
    portRangeHints[MAX_TIME].UpperBound = 100;
    portRangeHints[MANUAL_MODE].HintDescriptor = LADSPA_HINT_TOGGLED;
    portRangeHints[DATA_INPUT].HintDescriptor = 0;
    portRangeHints[DATA_OUTPUT].HintDescriptor = 0;
    descriptor.PortRangeHints = portRangeHints;
    
    descriptor.instantiate = instantiate;
    descriptor.connect_port = connect_port;
    descriptor.activate = activate;
    descriptor.run = run;
    descriptor.run_adding = 0;
    descriptor.set_run_adding_gain = 0;
    descriptor.deactivate = deactivate;
    descriptor.cleanup = cleanup;
}

Loader init;

} // namespace

// This is the function the client will use to get a handle
// to the plugin descriptor.
extern "C" {
const LADSPA_Descriptor*
ladspa_descriptor (unsigned long idx)
{
    if (idx > 0) {
       return 0;
    }
    return init.theDescriptor();
}
} // extern C
