/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/

/******************************************************************************
 * We keep, inside DataBlockSet, two vectors<struct iovec> and a
 * list<BlockInfo>. Each BlockInfo contains the information relative to the
 * block of data we want to manipulate: base address, size, type etc.  A "copy"
 * of part of this data is kept on the iovecs; if the blocks correspond to
 * input data, on the iovecIn; if it corresponds to output data, on the
 * iovecOut; if input/output, on both. At the same time, the BlockInfo keeps
 * the index numbers of the elements on the iovec with the data it is
 * responsible for.
 *
 * The sent data is made of two parts: the first half are the sizes of the
 * blocks of data of the second half. To do this, we keep a vector with the
 * sizes as the first element of the "send" vector. On the remote side, readv
 * reads this data and automatically "fixes" the size information of the struct
 * iovec vector being processed; so, when readv reaches the second half of the
 * struct iovec vector, all size information has already been updated. To do
 * this, the "receive" vector has, at the first half, struct iovec's that refer
 * to the sizes of the elements of the second half. For this to work, the size
 * data is sent with the receiver's endianness.
 *
 * The elements on the iovec's are kept on the same order as in the list,
 * that is, traversing the list searching for the input blocks will give us
 * the blocks in iovecIn on the same order; the same is valid for iovecOut.
 * Whenever we add or remove elements from the list we need to insert/remove
 * the corresponding elements on the iovecs and traverse the list to correct
 * the indexes kept inside the BlockInfos, which is only possible if
 * everything is kept on the same order.
 *
 * Finally, DataBlock is a smart pointer to a BlockInfo allowing the user to
 * change the block's characteristics: input/output, base address etc. It also
 * works as an iterator to traverse the DataBlockSet; this is possible because
 * it keeps an iterator that points to the list<BlockInfo> that belongs to the
 * DataBlockSet.
 ******************************************************************************/

// FIXME Right now, every block is treated as if "variableSize" was set to true.


using namespace std;

#include <dlowlat/masterslave/datablocks.h>

#include <algorithm>
#include <function.h>

#include <ladspa_sdk/ladspa.h>


namespace DLowLat {

/******************************************************************************
 *                     misc functions with file scope
 ******************************************************************************/

// Dumb byte-swapping functions, used to change the data endianness.
// Surely there's a standard, portable and faster way of doing this,
// but since I couldn't find it, I just coded it.
inline void
swap32 (void* ptr)
{
    unsigned long& value = *(static_cast<unsigned long*>(ptr));
    value = ((value & 0x000000ff) << 24) |
        ((value & 0x0000ff00) << 8) |
        ((value >> 8) & 0x0000ff00) |
        ((value >> 24) & 0x000000ff);
};

inline void
swap16 (void* ptr)
{
    unsigned short& value = *(static_cast<unsigned short*>(ptr));
    value = ((value & 0x00ff) << 8) |
        ((value >> 8) & 0x00ff);
};


/******************************************************************************
 *                          Object methods
 ******************************************************************************/

///////////// DataBlock methods ////////////

DataBlock::DataBlock (const BlockListPtr& block, DataBlockSet& set)
        : block (block)
{
    container = &set;
};

DataBlock
DataBlock::erase ()
{
    // This removes the element we point to, so we change ourselves to refer to
    // the next element. Also, we return a DataBlock that corresponds to the
    // next element too, like the standard list container.
    block = container->erase (block);
    return DataBlock (block, *container);
};

const DataBlock
DataBlock::operator++ (int) const
{
    DataBlock tmp (*this);
    ++tmp;
    return tmp;
};

const DataBlock
DataBlock::operator-- (int) const
{
    DataBlock tmp (*this);
    --tmp;
    return tmp;
};

DataBlock&
DataBlock::operator++ () {
    ++block;
    return *this;
};

DataBlock&
DataBlock::operator-- () {
    --block;
    return *this;
};

///////////// DataBlockSet methods /////////////

DataBlockSet::DataBlockSet ()
{
    // This is the block that contains the sizes of the other output blocks.
    iovecOut.push_back (struct iovec());
    iovecOut[0].iov_len = 0;
    iovecOut[0].iov_base = 0;
    firstInputIndex = 0;
};

DataBlock
DataBlockSet::newDataBlock ()
{
    // Create an empty block and return an iterator to it.
    blocks.push_back (BlockInfo());
    blocks.back().direction = NONE;
    return DataBlock (--blocks.end(), *this);
};

void
DataBlockSet::pop_back ()
{
    erase (--blocks.end());
};

void
DataBlockSet::prepareToSendData (Endianness remoteEndianness)
{
    // The client may have altered the size of the data to be
    // processed; copy the size information into the outputSizes
    // and the struct iovec vectors.
    list<BlockInfo>::iterator it, end (blocks.end());
    unsigned i = 1;
    for (it = blocks.begin(); it != end; ++it) {
        if (isOut (it->direction)) {
            iovecOut[it->indexOut].iov_len =
                *(it->numElems) * it->elementSize;
            outputSizes[i-1] = iovecOut[i].iov_len;
            ++i;
        }
    }

    // If needed, change the endianness of the outputSizes data, so that the
    // remote end receives this data with the correct endianness. This is
    // necessary because this data is used to update the iov_len for the
    // elements in the struct iovec array "inside" the call to readv(): after
    // readv() processes the first half of the array (by reading the sizes),
    // the size of the subsequent elements is correct and readv() will continue
    // the processing putting the data on the appropriate buffers.
    if (! isHostOrder(remoteEndianness)) {
        for (unsigned i = 0, tam = outputSizes.size(); i < tam; ++i) {
            swap32 (static_cast<void*> (&(outputSizes[0]) + i));
        }
    }
};

// fix the endianness of every element in the vector if necessary
// an calculate the number of blocks received from the remote end.
void
DataBlockSet::correctReceivedData (Endianness remoteEndianness)
{
    BlockListPtr it, end (blocks.end());

    if (! isHostOrder(remoteEndianness)) {
        for (it = blocks.begin(); it != end; ++it) {
            if (isIn (it->direction) && it->endiannessMatters) {
                toggleEndianness (iovecIn[it->indexIn], it->elementSize);
            }
        }
    }

    // We know the number of bytes received for each block; now let's
    // calculate the corresponding number of elements and make this
    // information accessible to the client. This data is already
    // in the correct endianness!
    for (it = blocks.begin(); it != end; ++it) {
        if (isIn (it->direction)) {
            *(it->numElems) = iovecIn[it->indexIn].iov_len / it->elementSize;
        }
    }
};

void
DataBlockSet::zeroAllInputs ()
{
    BlockListPtr it, end (blocks.end());
    for (it = blocks.begin(); it != end; ++it) {
        if (isIn (it->direction)) {
            struct iovec& theIovec (iovecIn[it->indexIn]);
            for (unsigned i = 0; i < theIovec.iov_len; ++i) {
                static_cast<char*>(theIovec.iov_base)[i] = 0;
            }
        }
    }
};

void
DataBlockSet::changeBlock (BlockListPtr block, Direction oldDirection)
{
    if (isIn (block->direction)) {
        if (! isIn (oldDirection)) {
            newInput(block);
        }
    } else if (isIn (oldDirection)) {
        eraseInput(block);
    }

    if (isOut (block->direction)) {
        if (! isOut (oldDirection)) {
            newOutput(block);
        }
    } else if (isOut (oldDirection)) {
        eraseOutput(block);
    }
};

BlockListPtr
DataBlockSet::erase (BlockListPtr block)
{
    if (isIn (block->direction)) {
        isIn (block->direction, false);
        eraseInput (block);
    }
    if (isOut (block->direction)) {
        isOut (block->direction, false);
        eraseOutput (block);
    }
    return blocks.erase (block);
};

void
DataBlockSet::newInput (BlockListPtr block)
{
    // Insert a new block *and* a block that refers to the size
    // of this new block; The first half of the vector contains
    // blocks that make reference to the sizes of the blocks on
    // the second half.

    // Find previous block that is an input; we will be inserted
    // in the struct iovec vector after it.
    bool firstInput = false;
    BlockListPtr prev = block;
    std::vector<struct iovec>::iterator it, newInputSize;
    while (prev != blocks.begin() && ! isIn ((--prev)->direction));

    if (prev == blocks.begin()) {
        if (block == blocks.begin() || !isIn (prev->direction)) {
            firstInput = true;
        }
    }
    if (firstInput) {
        it = iovecIn.begin();
        iovecIn.insert (it, struct iovec());
        it = iovecIn.begin();
        newInputSize = iovecIn.insert (it, struct iovec());
    } else {
        it = iovecIn.begin() + prev->indexIn + 1;
        iovecIn.insert (it, struct iovec());
        it = iovecIn.begin() + prev->indexIn - firstInputIndex + 1;
        newInputSize = iovecIn.insert (it, struct iovec());
    }
    newInputSize->iov_len = sizeof (size_t);
    ++firstInputIndex;
    realignInputs();
};

void
DataBlockSet::newOutput (BlockListPtr block)
{
    // Find previous block that is an output; we will be inserted
    // in the struct iovec vector after it.
    bool firstOutput = false;
    BlockListPtr prev = block;
    std::vector<struct iovec>::iterator it;
    std::vector<size_t>::iterator tmp = outputSizes.begin();
    while (block != blocks.begin() && ! isOut ((--block)->direction));

    if (prev == blocks.begin()) {
        if (block == blocks.begin() || !isOut (prev->direction)) {
            firstOutput = true;
        }
    }
    if (firstOutput) {
        it = iovecOut.begin();
        // We need to skip the first element, which contains the vector
        // of the sizes of the subsequent elements.
        iovecOut.insert (++it, struct iovec());
    } else {
        it = iovecOut.begin() + block->indexOut + 1;
        iovecOut.insert (it, struct iovec());
        tmp += block->indexOut;
    }

    outputSizes.insert (tmp, size_t());
    realignOutputs();
};


void
DataBlockSet::eraseInput (BlockListPtr block)
{
    iovecIn.erase (iovecIn.begin() + block->indexIn);
    iovecIn.erase (iovecIn.begin() + block->indexIn - firstInputIndex);
    --firstInputIndex;
    realignInputs();
};

void
DataBlockSet::eraseOutput (BlockListPtr block)
{
    iovecOut.erase (iovecOut.begin() + block->indexOut);
    outputSizes.erase (outputSizes.begin() + block->indexOut);
    realignOutputs();
};

void
DataBlockSet::realignInputs ()
{
    // Resizing a vector may cause reallocs, which
    // would mess the pointers.
    for (unsigned int i = 0; i < firstInputIndex; ++i) {
        iovecIn[i].iov_base = &(iovecIn[i + firstInputIndex].iov_len);
    }

    BlockListPtr ptr = blocks.begin();
    for (unsigned int i = firstInputIndex; i < iovecIn.size(); ++i) {
        while (! isIn(ptr->direction)) {
            ++ptr;
        }
        ptr->indexIn = i;
        ++ptr;
    }
};

void
DataBlockSet::realignOutputs ()
{
    // Resizing a vector may cause reallocs, which
    // would mess the pointers.
    iovecOut[0].iov_len = outputSizes.size() * sizeof (size_t);
    iovecOut[0].iov_base = &outputSizes[0];

    // Loop starts at i == 1 instead of 0 to skip the "sizes" block.
    BlockListPtr ptr = blocks.begin();
    for (unsigned int i = 1; i < iovecOut.size(); ++i) {
        while (! isOut(ptr->direction)) {
            ++ptr;
        }
        ptr->indexOut = i;
        ++ptr;
    }
};

inline void
DataBlockSet::toggleEndianness (struct iovec& theIovec, const size_t& elementSize)
{
    // FIXME Not 64-bit clean...
    // Aaargh! The attack of the casts! What did you expect...
    for (unsigned i = 0; i < theIovec.iov_len; i += elementSize) {
        if (elementSize == 4) {
            swap32 (static_cast<void*> (static_cast<char*>(theIovec.iov_base) + i));
        } else {
            // Same as above, but for 16 bits.
            swap16 (static_cast<void*> (static_cast<char*>(theIovec.iov_base) + i));
        }
    }
};

} // Namespace DLowLat
