#ifndef DLOWLAT_DATABLOCKS_H
#define DLOWLAT_DATABLOCKS_H

/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * These classes act as wrappers around struct iovec; they allow us to easily
 * pass around data pointers to different types of data without the need to
 * cast pointers to and from void and almost without having to worry about
 * endianness.
 *
 * To use them, we create a DataBlockSet (which is similar to a container
 * for DataBlocks) and ask for it to create new DataBlocks for us:
 *
 * DataBlock block (DataBlockSet.newDataBlock());
 *
 * Then we manipulate the DataBlock like so:
 *
 * block->setBlock (thePointer, numElems, IN/OUT, true);
 *
 * DataBlock may also be used as a bidirectional iterator to traverse
 * the DataBlockSet with the usual iterator syntax (it actually *is*
 * implemented as a list iterator).
 *
 * If a DataBlock is created with "variableSize" set to "true", the client must
 * check the current value for the "numElems" variable, which may be set if the
 * block size has changed.
 *
 * For each data type we want to be able to support in these classes, we need
 * to edit the "setBlock()" function to add the necessary typeid.
 *
 * WARNING! Don't try to create a mechanism for the type "void*"; this might
 * allow any type for which no typeid exists to be "seen" as a void (thanks to
 * implicit conversions); that would probably *not* give the expected results.
 ******************************************************************************/


#include <dlowlat/masterslave/endianness.h>

// Needed because of the setBlock() template.
#include <typeinfo>

// FIXME!! gcc 2.95 doesn't know about C++ numeric limits
#include <climits>

#include <sys/uio.h> // struct iovec

#include <vector>
#include <list>
#include <iostream> // cerr


/******************************************************************************
 *                           Type declarations
 ******************************************************************************/

// Right now, this is the only "weird" type (that is, a type that should be
// wrapped which needs to be declared) around here:
typedef float LADSPA_Data;

namespace DLowLat {

// A DataBlock may be for data to be sent, received, or both
// FIXME: Maybe this should be a class
enum Direction
{
    NONE = 0,
    IN = 1,
    OUT = 2,
    INOUT = 3
};

inline void
isIn (Direction& d, bool val)
{
    if (val) {
        // FIXME These casts shouldn't need to be here
        static_cast<int>(d) |= IN;
    } else {
        static_cast<int>(d) &= ~IN;
    }
}

inline void
isOut (Direction& d, bool val)
{
    if (val) {
        // FIXME These casts shouldn't need to be here
        static_cast<int>(d) |= OUT;
    } else {
        static_cast<int>(d) &= ~OUT;
    }
}

inline bool
isIn (Direction d)
{
    return d & IN;
}

inline bool
isOut (Direction d)
{
    return d & OUT;
}


/******************************************************************************
 *                             Data Structure
 ******************************************************************************/

// These declarations are of interest for the internals of these classes only

struct BlockInfo {
    void* theData; // Block that we point to
    // Number of elements on the block; a pointer because we may change it.
    size_t* numElems;
    // Size in bytes for an instance of the wrapped data.
    size_t elementSize;
    bool endiannessMatters; // for chars it doesn't.
    bool variableSize;
    Direction direction;
    size_t indexIn, indexOut;
};

class DataBlockSet;
typedef std::list<BlockInfo>::iterator BlockListPtr;


/******************************************************************************
 *           A DataBlock is a "pseudo-wrapper" around a block of data
 ******************************************************************************/

class DataBlock {
public:
    // Each DataBlock must hold a reference to the DataBlockSet to which it
    // belongs in order to be able to ask for the in and out struct iovec's as
    // needed. It is expected that only DataBlockSet creates new objects of
    // this type, passing a reference to themselves to the constructor.
    DataBlock (const BlockListPtr&, DataBlockSet&);

    // The most important method; here the client registers a block of
    // data on the DataBlockSet that is to be manipulated.
    template <class T>
    void setBlock (T* ptr, size_t& numElems, Direction inout, bool variableSize);

    // FIXME Right now clients use these besides setBlock,
    // but they might be removed.
    inline void blockAddress (void*);
    inline void numElems (size_t&);
    inline void direction (Direction);
    inline void variableSize (bool);

    DataBlock erase ();
    inline DataBlock* operator-> ();
    inline DataBlock& operator* ();
    DataBlock& operator++ ();
    DataBlock& operator-- ();
    const DataBlock operator++ (int) const;
    const DataBlock operator-- (int) const;

private:
    DataBlockSet* container;
    BlockListPtr block;
};


/******************************************************************************
 *           This is a collection of references to blocks of data
 *                and a factory of corresponding DataBlocks
 ******************************************************************************/

class DataBlockSet {
public:
    DataBlockSet ();

    // These are used by the clients
    DataBlock newDataBlock ();
    void pop_back ();

    // These are used by the Master and Slave classes
    inline struct iovec* inputs ();
    inline struct iovec* outputs ();
    inline size_t numInputs () const;
    inline size_t numOutputs () const;
    void prepareToSendData (const Endianness remoteEndianness);
    void correctReceivedData (const Endianness remoteEndianness);
    void zeroAllInputs ();

    // These are used by DataBlock
    void changeBlock (BlockListPtr block, Direction oldDirection);
    BlockListPtr erase (BlockListPtr);
private:
    void newInput (BlockListPtr);
    void newOutput (BlockListPtr);
    void eraseInput (BlockListPtr);
    void eraseOutput (BlockListPtr);
    void realignInputs ();
    void realignOutputs ();
    inline void toggleEndianness (struct iovec&, const size_t&);

    std::vector<struct iovec> iovecIn, iovecOut;
    // This is a list and not a vector or something else so that the
    // iterators we provide to the users (inside the DataBlock objects)
    // aren't invalidated by insertions/deletions of elements inside
    // this container.
    std::list<BlockInfo> blocks;
    std::vector<size_t> outputSizes;
    EndiannessChecker isHostOrder;
    size_t firstInputIndex;
};


/******************************************************************************
 *                               templates
 ******************************************************************************/

// Add your own typeids here. Note that this only works for data that is
// linearly arranged in memory, starting at the location indicated by the
// pointer. For complex types, you probably need to "serialize" and
// "de-serialize" the object's state in memory and change the
// "toggleEndianness" method; but probably complex classes don't belong here
// anyway.

template <class T>
void
DataBlock::setBlock (T* ptr, size_t& numElems, Direction inout, bool variableSize)
{
    Direction oldDirection = block->direction;
    if (
        typeid (T) == typeid (int) ||
        typeid (T) == typeid (unsigned int) ||
        typeid (T) == typeid (short) ||
        typeid (T) == typeid (unsigned short) ||
        typeid (T) == typeid (long) ||
        typeid (T) == typeid (float) ||
        typeid (T) == typeid (unsigned long)
    ) {
        block->endiannessMatters = true;
    } else if (
        typeid (T) == typeid (char) ||
        typeid (T) == typeid (unsigned char)
    ) {
        block->endiannessMatters = false;
    } else {
        // FIXME! Error checking
        std::cerr << "DLowLat::DataBlock class: Unknown type!" << std::endl;
        exit (1);
    }
    
    block->elementSize = sizeof (T);
    direction (inout);
    block->theData = static_cast<void*>(ptr);
    this->variableSize (variableSize);
    block->numElems = &numElems;
    // Invalidate the indexes, these will have to be redefined by the container
    // FIXME gcc 2.95 doesn't know how to handle this
    //block->indexIn = block->indexOut = std::numeric_limits<size_t>::max();
    block->indexIn = block->indexOut = INT_MAX;
    container->changeBlock (block, oldDirection);
    blockAddress (static_cast<void*>(ptr));
    this->numElems (numElems);
};


/******************************************************************************
 *                               inlines 
 ******************************************************************************/

////////////// DataBlock ///////////////

inline void
DataBlock::blockAddress (void* ptr)
{
    block->theData = ptr;
    if (isIn (block->direction)) {
        container->inputs()[block->indexIn].iov_base = ptr;
    }
    if (isOut (block->direction)) {
        container->outputs()[block->indexOut].iov_base = ptr;
    }
};

inline void
DataBlock::numElems (size_t& numElems)
{
    block->numElems = &numElems;
    if (isIn (block->direction)) {
        container->inputs()[block->indexIn].iov_len = numElems * block->elementSize;
    }
    if (isOut (block->direction)) {
        container->outputs()[block->indexOut].iov_len = numElems * block->elementSize;
    }
};

inline void
DataBlock::direction (Direction d)
{
    block->direction = d;
};

inline void
DataBlock::variableSize (bool variableSize)
{
    block->variableSize = variableSize;
};


inline DataBlock*
DataBlock::operator-> ()
{
    return this;
};

inline DataBlock&
DataBlock::operator* ()
{
    return *this;
};


////////////// DataBlockSet //////////////

inline struct iovec*
DataBlockSet::inputs ()
{
    return &iovecIn[0];
};

inline struct iovec*
DataBlockSet::outputs ()
{
    return &iovecOut[0];
};

inline size_t
DataBlockSet::numInputs () const
{
    return iovecIn.size();
};

inline size_t
DataBlockSet::numOutputs () const
{
    return iovecOut.size();
};

} // Namespace DLowLat

# endif // DLOWLAT_DATABLOCKS_H
