/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


using namespace std;

#include <dlowlat/masterslave/masterslave.h>

#include <miscutils/cronometer.h>

#include <errno.h>


namespace DLowLat {


/////////////// Master methods /////////////

Master::Master (UdpWrite& tmpWriter, UdpRead& tmpReader,
        Endianness remoteEndianness, int windowSize = 0) :
            writer (tmpWriter), reader (tmpReader),
            counterOutBlock (theSet.newDataBlock()),
            counterInBlock (theSet.newDataBlock())
{
    counterSize = 1;
    this->remoteEndianness = remoteEndianness;
    theWindowSize = windowSize;
    skipFrames = windowSize;
    counterOut = counterIn = 0;
    counterOutBlock->setBlock (&counterOut, counterSize, OUT, false);
    counterInBlock->setBlock (&counterIn, counterSize, IN, false);
};

void
Master::process ()
{
    ++counterOut;
    theSet.prepareToSendData (remoteEndianness);
    writer (theSet.outputs(), theSet.numOutputs());

    if (skipFrames > 0) {
        // We need to wait more cycles before we
        // actually have data to read.
        --skipFrames;
        theSet.zeroAllInputs ();
    } else {
        // Read the data
        do {
            Cronometer cron;
            cron.start();
            while (reader(theSet.inputs(), theSet.numInputs()) < 0 &&
                                                            errno == EAGAIN) {
                // Busy-wait if needed
                if (cron.elapsed () > MAX_SOCKET_POLL_TIME) {
                    // this I/O messes the timing completely,
                    // but we're using I/O because the timing
                    // is already messed up!
                    // FIXME: This should raise an exception
                    cerr << "socket timeout!!!" << endl;
                    // We're doomed; let's get out of here!
                    return;
                }
            }
        // Read until we receive the "right" frame. Note that the
        // counters overflow, since they are chars, but that's ok.
        } while (!(static_cast<unsigned char>(counterIn + theWindowSize) ==
                                                                counterOut));

        // Data read ok! Let's fix the endianness etc.
        theSet.correctReceivedData (remoteEndianness);
    }
};


////////////// Slave methods /////////////

Slave::Slave (UdpWrite& tmpWriter, UdpRead& tmpReader,
        Endianness remoteEndianness, void f(void*)) :
            writer (tmpWriter), reader (tmpReader),
            counterBlock (theSet.newDataBlock())
{
    counterSize = 1;
    this->remoteEndianness = remoteEndianness;
    counter = 0;
    counterBlock->setBlock (&counter, counterSize, INOUT, false);
    processFunction = f;
};

void
Slave::process(void* ptr)
{
    reader (theSet.inputs(), theSet.numInputs());
    theSet.correctReceivedData (remoteEndianness);
    processFunction (ptr);
    theSet.prepareToSendData (remoteEndianness);
    writer (theSet.outputs(), theSet.numOutputs());
};

} // Namespace DLowLat
