#ifndef DLOWLAT_MASTERSLAVE_H
#define DLOWLAT_MASTERSLAVE_H

/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * These classes function as a mechanism to allow the remote processing of data
 * in realtime; that is, they provide a synchronous or "pseudo-syncrhonous"
 * mode of operation (defined by the window size) where data existing in one
 * machine can be sent to another, processed and sent back as fast as the
 * low-level link permits. For example, a multimedia application may process
 * some of the data remotely using these classes. Pointers to the data to be
 * sent are stored on the DataBlockSet and, after a call to the process()
 * method of the Master class, can be considered to be processed. On the remote
 * end, the process() method from the Slave class is called continuously on an
 * infinite loop; at each call, it calls a function registered with it, passing
 * to it the DataBlockSet with the data received. This function is supposed to
 * process the data and then return. After it returns, the Slave class sends
 * the processed data to the sender machine.
 ******************************************************************************/


#include <dlowlat/transport/ip/udp/sockets.h>

#include <dlowlat/masterslave/datablocks.h>

#include <dlowlat/masterslave/endianness.h>


enum {
    // Maximum time (in microseconds) that we keep on retrying on a
    // non-blocking socket, default is 500ms (which is a *lot*); needed because
    // we may run with priority SCHED_FIFO and could lock the machine in an
    // infinite loop.
    MAX_SOCKET_POLL_TIME = 500000
};

namespace DLowLat {


/******************************************************************************
 *                           the master class
 ******************************************************************************/
    
class Master {

public:
    Master (UdpWrite& tmpWriter, UdpRead& tmpReader,
            Endianness remoteEndianness, int windowSize = 0);

    void process ();

    // So that we can change and query the window size during operation
    inline void windowSize (unsigned int); // 0, 1 or 2
    inline unsigned int windowSize () const;

    inline void reset ();

    inline DataBlock newDataBlock();

    IPAddress peerAddress () const { return writer.address(); };

private:
    UdpWrite writer;
    UdpRead reader;
    DataBlockSet theSet;
    Endianness remoteEndianness;

    unsigned char theWindowSize;
    // These are chars so that we don't have to worry about endianness; there's
    // code to deal with the fact that they will overflow periodically.
    unsigned char counterOut, counterIn;
    DataBlock counterOutBlock, counterInBlock;

    // When processing starts or the window size changes, we may need to skip
    // frames to re-sync. This is signaled by this variable being bigger than
    // zero.
    unsigned int skipFrames;

    // Although the size of the counter is fixed, DataBlock needs to take a
    // pointer to this size.
    size_t counterSize;
};


/******************************************************************************
 *                          the slave class
 ******************************************************************************/

// This class calls the registered function f(void*) periodically. This
// function may need no parameters, but if for any reason we need to send
// something (say we want a variable object to process the data; in this case,
// we need to pass the object to the callback), that's what "void*" is there
// for.

class Slave {

public:
    Slave (UdpWrite& tmpWriter, UdpRead& tmpReader,
            Endianness remoteEndianness, void f(void*));
    // Read the remote data, call the registered callback function and write
    // the data out.
    void process (void*);

    inline DataBlock newDataBlock();

private:
    Endianness remoteEndianness;
    UdpWrite writer;
    UdpRead reader;
    DataBlockSet theSet;

    // This is a char so that we don't have to worry about endianness; there's
    // code to deal with the fact that it will overflow periodically.
    unsigned char counter;
    DataBlock counterBlock;
    // Although the size of the counter is fixed, DataBlock needs to take a
    // pointer to this size.
    size_t counterSize;

    // "processFunction" is a pointer to the callback function
    void (*processFunction)(void*);
};


/******************************************************************************
 *                               inlines
 ******************************************************************************/

inline void
Master::windowSize(unsigned int i)
{
    if (i > theWindowSize) {
        skipFrames = i - theWindowSize;
    }
    theWindowSize = i;
};

inline unsigned int
Master::windowSize () const
{
    return theWindowSize;
};

inline void
Master::reset ()
{
    reader.flush();
    skipFrames = theWindowSize;
};

inline DataBlock
Master::newDataBlock ()
{
    return theSet.newDataBlock();
};

inline DataBlock
Slave::newDataBlock ()
{
    return theSet.newDataBlock();
};

} // Namespace DLowLat

#endif // DLOWLAT_MASTERSLAVE_H
