/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


using namespace std;

#include <dlowlat/transport/ip/ipaddress.h>

#include <miscutils/hostname.h> // my own version of getHostname()

#include <string>
#include <sstream>
#include <iostream>

// inet_ntop, getipnodebyname, resolver etc.
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <string.h> // strcmp

namespace DLowLat {

// What chars are possible in a numeric IP?
static const char* numeric = "1234567890.";

IPAddress::IPAddress (const string& hostname = "0.0.0.0:0")
{
    // "Clear" the structure
    memset (&address, 0, sizeof (sockaddr_in));
    address.sin_family = AF_INET;
    port (0); // In case the user didn't inform a port number
    host (hostname);
};

string
IPAddress::host () const
{
    char buf[INET6_ADDRSTRLEN] = {0};
    buf[0] = '\0';
    inet_ntop(address.sin_family, static_cast<const void*>
            (&address.sin_addr), buf, INET6_ADDRSTRLEN);
    if (strcmp (buf, "0.0.0.0") == 0) {
        // INADDR_ANY is not a very good answer... let's return one of
        // this host's IPs.
        return thisHostsIP ();
    } // else
    return buf;
};

string
IPAddress::thisHostsIP () const
{
    // AAARGH! There's no getipnodebyname & friends on the linux includes,
    // although the manpages exist
    // int error;
    //struct hostent* hostdata = getipnodebyname (getHostname().c_str(), AF_INET, 0, &error);
    struct hostent* hostdata = gethostbyname (getHostname().c_str());

    if (hostdata == 0) {
        // FIXME
        cerr << "ipaddress.cc, class IPAddress: ";
        cerr << "unable to resolve hostname; ";
        cerr << "no network information available" << endl;
        exit (1);
    }

    char buf[INET6_ADDRSTRLEN] = {0};
    buf[0] = '\0';
    if (inet_ntop (address.sin_family, hostdata->h_addr_list[0], buf, INET6_ADDRSTRLEN) == "127.0.0.1") {
        // Localhost is not a good answer; we'll only use it if the
        // machine doesn't have any other address
        if (! hostdata->h_addr_list[1] == 0) {
            buf[0] = '\0';
            inet_ntop (address.sin_family, hostdata->h_addr_list[1], buf, INET6_ADDRSTRLEN);
        }
    }
    // This should be here if we used getipnodebyname()
    //freehostent (hostdata);

    return buf;
};

uint16_t
IPAddress::port () const
{
    return ntohs(address.sin_port);
};

void
IPAddress::host (const string& hostname)
{
    string::size_type colon = hostname.find (':');
    if (colon != string::npos) {
        port (hostname.substr(colon + 1));
    }
    string hostAddress = resolveHostname (hostname.substr (0, colon));

    if (hostAddress == "0.0.0.0") {
        address.sin_addr.s_addr = htonl (INADDR_ANY);
    } else {
        inet_pton (address.sin_family, hostAddress.c_str(), &address.sin_addr);
    }
};

void
IPAddress::port (const uint16_t portnum)
{
    address.sin_port = htons (portnum);
};

void
IPAddress::port (const string& portnum)
{
    if (portnum.find_first_not_of (numeric) != string::npos) {
        // This is a port name; we need to convert to a port number.
        struct servent* sptr = getservbyname (portnum.c_str(), "udp");
        port (ntohs(sptr->s_port));
        // No need to free() sptr
    } else {
        istringstream s (portnum);
        uint16_t i;
        s >> i;
        port (i);
    }
};

string
IPAddress::resolveHostname (const string& hostname)
{
    if (hostname.find_first_not_of (numeric) != string::npos) {
        // This is not a numeric IP, we need to resolve
        // FIXME gethostbyname is deprecated; use getipnodebyname() instead
        struct hostent* hptr = gethostbyname(hostname.c_str());
        char buf[INET6_ADDRSTRLEN] = {0};
        buf[0] = '\0';
        inet_ntop(hptr->h_addrtype, hptr->h_addr_list[0],
                        buf, INET6_ADDRSTRLEN);
        // No need to free (hptr);
        return buf;
    }
    // else, the hostname was already numeric
    return hostname;
};

} // namespace
