#ifndef DLOWLAT_IPADDRESS_H
#define DLOWLAT_IPADDRESS_H

/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * A small class for dealing with IP addresses in a manageable form
 * 
 * This class wraps the standard "struct sockaddr_in" and automatically
 * converts hostnames or hostnumbers and portnames or portnumbers to
 * a sockaddr_in useable by standard system calls.
 ******************************************************************************/

// Files that include this one may also need to include:
// <string> - C++-style strings
// <sys/socket.h>


#include <sstream>
#include <string.h> // memset
#include <netinet/in.h> // sockaddr etc.

namespace DLowLat {

class IPAddress {
public:
    // We may receive a number or a string as the port number
    template <class T>
    IPAddress (const std::string& hostname, const T& portnum);
    IPAddress (const std::string& hostname = "0.0.0.0:0");

    inline struct sockaddr* sockaddr();
    inline socklen_t size() const;

    std::string host() const;
    void host(const std::string& hostname); // may include the port too

    uint16_t port() const;
    void port(const uint16_t portnum);
    void port(const std::string& portnum);

private:
    struct sockaddr_in address;
    std::string resolveHostname (const std::string&);
    std::string thisHostsIP () const;
};


/******************************************************************************
 *                                Templates
 ******************************************************************************/

// "Stringify" an Address to be inserted into/extracted from a stream.
template <typename T>
inline
T&
operator << (T& stream, const IPAddress& address)
{
    std::ostringstream s; 
    s << address.host() << ":" << address.port();
    stream << s.str();
    return stream;
};

template <typename T>
inline
T&
operator >> (T& stream, IPAddress& address)
{
    std::string s;
    stream >> s;
    address.host(s);
    return stream;
};

template <class T>
IPAddress::IPAddress (const std::string& hostname, const T& portnum)
{
    // "Clear" the structure
    memset (&address, 0, sizeof (sockaddr_in));
    address.sin_family = AF_INET;

    std::ostringstream buf;
    buf << portnum;

    port (buf.str());
    host (hostname);
};


/******************************************************************************
 *                                 Inlines
 ******************************************************************************/

inline struct sockaddr*
IPAddress::sockaddr()
{
    // Hopefully no one will notice we're passing a handler to
    // private data ;-)
    return reinterpret_cast<struct sockaddr*> (&address);
};

inline socklen_t
IPAddress::size() const
{
    return sizeof (sockaddr_in);
};

} // namespace

#endif // DLOWLAT_IPADDRESS_H
