#ifndef DLOWLAT_CPUSTATS_H
#define DLOWLAT_CPUSTATS_H

/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/

/******************************************************************************
 * This class gathers statistic information on CPU and network usage from the
 * Linux /proc pseudo-filesystem. It works in a way similar to a cronometer:
 * the user starts it and, after some time, stops it. The statistics for this
 * period are then available to the user. Everything is inlined just to avoid
 * creating another .cc file.
 ******************************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

class CPUStats
{
public:
    inline CPUStats();
    inline ~CPUStats();
    inline void start();
    inline void stop();
    inline int userTimePercent();
    inline int systemTimePercent();
    inline int idleTimePercent();
    inline double collisionsPercent();
private:
    char buf[1024];
    int proc_stat, proc_net_dev;
    unsigned long initial_utime, initial_stime, initial_itime,
                    final_utime, final_stime, final_itime, initial_packets,
                    initial_collisions, final_packets, final_collisions;
};

inline
CPUStats::CPUStats()
{
    // FIXME These should be set as constants on the beginning of this file
    proc_stat = open ("/proc/stat", O_RDONLY);
    proc_net_dev = open ("/proc/net/dev", O_RDONLY);
};

inline
CPUStats::~CPUStats()
{
    close (proc_stat);
    close (proc_net_dev);
};

inline void
CPUStats::start ()
{
    lseek (proc_stat, 5, SEEK_SET);
    read (proc_stat, buf, 100);
    buf[100] = '\0';
    char *next;

    initial_utime = strtol (buf, &next, 0);
    initial_utime += strtol (next, &next, 0);
    initial_stime = strtol (next, &next, 0);
    initial_itime = strtol (next, &next, 0);

    lseek (proc_net_dev, 0, SEEK_SET);
    read (proc_net_dev, buf, 1023);
    buf[1023] = '\0';
    
    next = strstr (buf, "eth0");
    next += 5;
    for (int i = 0; i < 9; ++i) {
        strtol (next, &next, 0);
    }
    initial_packets = strtol (next, &next, 0);
    strtol (next, &next, 0);
    strtol (next, &next, 0);
    strtol (next, &next, 0);
    initial_collisions = strtol (next, &next, 0);
};

inline void
CPUStats::stop ()
{
    lseek (proc_stat, 5, SEEK_SET);
    read (proc_stat, buf, 100);
    buf[100] = '\0';
    char *next;

    final_utime = strtol (buf, &next, 0);
    final_utime += strtol (next, &next, 0);
    final_stime = strtol (next, &next, 0);
    final_itime = strtol (next, &next, 0);

    lseek (proc_net_dev, 0, SEEK_SET);
    read (proc_net_dev, buf, 1023);
    buf[1023] = '\0';
    
    next = strstr (buf, "eth0");
    next += 5;
    for (int i = 0; i < 9; ++i) {
        strtol (next, &next, 0);
    }
    final_packets = strtol (next, &next, 0);
    strtol (next, &next, 0);
    strtol (next, &next, 0);
    strtol (next, &next, 0);
    final_collisions = strtol (next, &next, 0);
};

inline int
CPUStats::userTimePercent()
{
    unsigned long totalTime = (final_utime + final_stime + final_itime) -
                            (initial_utime + initial_stime + initial_itime);
    return 100 * (final_utime - initial_utime) /totalTime;
};

inline int
CPUStats::systemTimePercent()
{
    unsigned long totalTime = (final_utime + final_stime + final_itime) -
                            (initial_utime + initial_stime + initial_itime);
    return 100 * (final_stime - initial_stime) /totalTime;
};

inline int
CPUStats::idleTimePercent()
{
    unsigned long totalTime = (final_utime + final_stime + final_itime) -
                            (initial_utime + initial_stime + initial_itime);
    return 100 * (final_itime - initial_itime) /totalTime;
};

inline double
CPUStats::collisionsPercent()
{
    return 100.0 * static_cast<double>(final_collisions - initial_collisions)
                        / static_cast<double>(final_packets - initial_packets);
}

/* 
 This is the way to get user and system time for the current process:

    char buf[1024];
    int i;
    unsigned long u, utime, stime;
    char c;
    self = fopen ("/proc/self/stat", "r");

    fscanf (self, "%d %s %c %d %d %d %d %d %u %u %u %u %u %u %u",
            &i, buf, &c, &i, &i, &i, &i, &i, &u, &u, &u, &u, &u, &utime, &stime);

 */

#endif // DLOWLAT_CPUSTATS_H
