#ifndef DLOWLAT_CRONOMETER_H
#define DLOWLAT_CRONOMETER_H

/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * A simple cronometer class; everything is inlined here just because
 * I'm lazy to add yet another file for compiling/linking in the Makefile
 *
 * If possible, you should #define USE_PENTIUM_TIMER, which is cheaper
 * and has more resolution.
 ******************************************************************************/


#include <sys/time.h>
#include <time.h>
#include <limits.h> // ULONG_MAX; some day we will use C++ limits instead

#include <fstream>
#include <string>
#include <sstream>

class Cronometer {
public:
    inline Cronometer ();
    inline void start ();
    inline void stop ();
    inline void resume ();
    inline unsigned long elapsed (); // in microseconds
    inline void busyWait (unsigned long sleepTime); // in microseconds
private:
    inline void calculateCPUMHz();
    inline void setStartTime();
    inline long fromStartToNow();
    struct timeval start_time, stop_time;
    unsigned long long startCycle, stopCycle;
    long accumulatedTime;
    double cpuMHz;
    bool running;
};


////////////////////////////////////////////////////////////////////////////////
//                                inlines
////////////////////////////////////////////////////////////////////////////////

inline
Cronometer::Cronometer ()
{
    calculateCPUMHz ();
    // Let's guarantee sane values for the variables,
    // even if we expect that start() is always called.
    start();
    running = false;
};

inline void
Cronometer::start ()
{
    accumulatedTime = 0;
    setStartTime ();
    running = true;
};

inline void
Cronometer::stop ()
{
    accumulatedTime = elapsed();
    running = false;
};

inline void
Cronometer::resume ()
{
    if (!running) {
        setStartTime();
        running = true;
    }
};

inline unsigned long
Cronometer::elapsed ()
{
    if (! running) {
        return accumulatedTime;
    } else {
        return accumulatedTime + fromStartToNow();
    }
};

inline void
Cronometer::busyWait (unsigned long sleepTime)
{
    // FIXME gcc 2.95 doesn't understand C++-style limits
    if (sleepTime >= ULONG_MAX/2 + 1) {
        // 35 minutes busy-waiting? Ha! Someone screwed up and came up
        // with a negative value, which backfired since this is unsigned.
        // Let's just get out of here fast!
        return;
    }
    start();
    while (sleepTime > elapsed()) { };
};        
  

// Here are the architecture-dependent #ifdefs

inline void
Cronometer::calculateCPUMHz ()
{
    cpuMHz = 0.0;
#ifdef USE_PENTIUM_TIMER
    // Find the CPU clock of this machine
    // FIXME: No error checking whatsoever!
    // FIXME: these names should be constants defined elsewhere
    std::ifstream cpuinfo ("/proc/cpuinfo");
    std::string cpuline ("cpu MHz");
    std::string line;
    while (cpuinfo.peek() != EOF) {
        getline (cpuinfo, line);
        // Aaaargh! Gcc 2.95 implements string::compare() differently from
        // the spec...
        //if (cpuline.compare (0, cpuline.size(), line) == 0) {
        if (cpuline.compare (line,0, cpuline.size()) == 0) {
            std::istringstream s (line.substr (line.find(':') + 1));
            s >> cpuMHz;
            break;
        }
    }
#endif
}

inline void
Cronometer::setStartTime ()
{
#ifdef USE_PENTIUM_TIMER
    __asm__ volatile (".byte 0x0f, 0x31" : "=A" (startCycle));
#else
    gettimeofday (&start_time, 0);
#endif
};

inline long
Cronometer::fromStartToNow ()
{
#ifdef USE_PENTIUM_TIMER
    __asm__ volatile (".byte 0x0f, 0x31" : "=A" (stopCycle));
    return static_cast<long>((stopCycle - startCycle) / cpuMHz);
#else
    gettimeofday (&stop_time, 0);
    return stop_time.tv_usec - start_time.tv_usec +
        (stop_time.tv_sec - start_time.tv_sec) * 1000000;
#endif
};

#endif // DLOWLAT_CRONOMETER_H
