/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


#include <sched.h> // realtime scheduling

#include <iostream> // cerr
#include <functional>
#include <algorithm>

#include <dlowlat/transport/ip/ipaddress.h>
#include <dladspa/container.h>
#include <dlowlat/transport/ip/udp/sockets.h>
#include <dlowlat/masterslave/masterslave.h>

#ifdef ORB_MICO
#include <plugin_factory/plugin_factory.h>
#include <mico/CosNaming.h>
#elif defined (ORB_OMNIORB)
#include <plugin_factory/plugin_factory.hh>
#endif

using namespace DLADSPA;
using namespace DLowLat;
using namespace std;

#include <plugin_factory/servant.h>

// Private stuff
namespace {


/******************************************************************************
 *                            Auxiliary stuff
 ******************************************************************************/

// The callback function that will be called by the slave driver
void registeredProcessFunction (void* ptr)
{
    static_cast<DLADSPA::Container*> (ptr)->process ();
};

// Can't use references here, why???
bool
isTheAddress (pair<IPAddress, pthread_t*> thePair, IPAddressSpec theAddress)
{
    return thePair.first.port() == theAddress.port &&
            strcmp (thePair.first.host().c_str(), theAddress.host) == 0;
};

// Needed forward declaration
void* pluginThread (void*);

pthread_t*
createThread (void* ptr)
{
    // Prepare attributes for the new thread
    pthread_attr_t attr;
    pthread_attr_init (&attr);
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);

    // Prepare new thread for real-time operation
    struct sched_param priority;
    priority.sched_priority = 98;
    pthread_attr_setschedpolicy (&attr, SCHED_FIFO);
    pthread_attr_setschedparam (&attr, &priority);
    pthread_attr_setinheritsched (&attr, PTHREAD_EXPLICIT_SCHED);
    pthread_attr_setscope (&attr, PTHREAD_SCOPE_PROCESS);

    // Create the new thread
    pthread_t* threadId = new pthread_t;
    pthread_create (threadId, &attr, pluginThread, ptr);
    return threadId;
};

/******************************************************************************
 *      The stuff that will actually run the plugin on a separate thread
 ******************************************************************************/

struct PluginThreadData
{
    const char* pluginDescriptor;
    DLowLat::UdpRead* reader;
    DLowLat::UdpWrite* writer;
    unsigned long sampleRate;
    DLowLat::Slave* transceiver;
    DLADSPA::Container* container;
};

void
cleanupPluginThread(void* ptr)
{
    struct PluginThreadData* threadData =
                                static_cast<struct PluginThreadData*> (ptr);
  
    delete threadData->container;
    delete threadData->transceiver;
    delete threadData->reader;
    delete threadData->writer;
    delete threadData;
};

void*
pluginThread (void* ptr)
{
    struct PluginThreadData* threadData =
                            static_cast<struct PluginThreadData*> (ptr);
    pthread_cleanup_push (cleanupPluginThread, ptr);
    threadData->transceiver = new Slave (*(threadData->writer),
            *(threadData->reader), LITTLEENDIAN, registeredProcessFunction);
    threadData->container = new Container (*(threadData->transceiver),
                    threadData->pluginDescriptor, threadData->sampleRate);
    int i;
    while (true) {
        threadData->transceiver->process (static_cast<void*>(threadData->container));
        // Maybe we received a cancellation request; while the standard says
        // read() and write() are cancellation points (Butenhof, p.147),
        // apparently there's a bug somewhere so they actually are not.
        // 100 is just some reasonable value to avoid a large overhead.
        if (++i > 100) {
            i = 0;
            pthread_testcancel();
        }
    }
    pthread_cleanup_pop (1);
    return 0;
};


}; // closes the unnamed namespace


/******************************************************************************
 *                 Implementation of PluginFactoryImp
 ******************************************************************************/

IPAddressSpec*
PluginFactoryImp::newPlugin (const IPAddressSpec& clientAddress,
                    const char* pluginDescriptor, SampleRate rate)
                    throw (CORBA::SystemException)
{
    // Prepare the connection
    IPAddress peerIP (string (clientAddress.host), clientAddress.port);
    IPAddress myIP;
    struct PluginThreadData* threadData = new PluginThreadData;
    threadData->reader = new UdpRead (myIP);
    threadData->writer = new UdpWrite (peerIP);
    threadData->sampleRate = rate;
    threadData->pluginDescriptor = pluginDescriptor;

    // Update myIP to the actual bind()'ed address.
    myIP = threadData->reader->address();

    // Create the new thread
    pthread_t* thread = createThread (static_cast<void*>(threadData));

    // Keep data on currently working threads and
    // return relevant data for client
    openContainers.push_back (make_pair (myIP, thread));

    // Prepare return value
    IPAddressSpec* returnAddress = new IPAddressSpec;
    returnAddress->host = CORBA::string_dup (myIP.host().c_str());
    returnAddress->port = myIP.port();
    return returnAddress;
};

void
PluginFactoryImp::deletePlugin (const IPAddressSpec& serverAddress)
                                    throw (CORBA::SystemException)
{
    list<pair<IPAddress, pthread_t*> >::iterator it;
    // I want to be able to define a lambda function! ;-)
    it = find_if (openContainers.begin(), openContainers.end(),
            bind2nd (ptr_fun(isTheAddress), serverAddress));

    if (it == openContainers.end()) {
        // FIXME!!!
        cerr << "Invalid server!!!" << endl;
        exit (1);
    } else {
        pthread_cancel (*(it->second));
        delete it->second;
        openContainers.erase (it);
    }
};

PluginFactoryImp::~PluginFactoryImp()
{
    list<pair<IPAddress, pthread_t*> >::iterator it;
    for (it = openContainers.begin(); it != openContainers.end(); ++it) {
       pthread_cancel (*(it->second));
       delete it->second;
    }
};
