/******************************************************************************
 * Copyright (C) 2003 - Nelson Posse Lago
 *
 * This file is part of DLADSPA, a system for distributed processing
 * of LADSPA plugins.
 *
 * DLADSPA is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * DLADSPA is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 * or just check their website at http://www.fsf.org
 ******************************************************************************/


/******************************************************************************
 * This is the server process that creates a CORBA object and registers it
 * on the CORBA name service. This object is a factory that creates pprocessing
 * objects as requested by the CORBA clients. The processing object is not a
 * CORBA object; it communicates with its peer directly. It should be very
 * easy to use this for other processor types besides LADSPA; inherit from
 * FactoryManager (just like DLADSPAFactoryManager), implement the adequate
 * defineCorbaNames() and startServant() methods, call startup() on the
 * constructor and shutdown() on the destructor and change the object that is
 * created on main() to be of the type you defined.
 ******************************************************************************/

// Signal/wait
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/mman.h> // mlockall

#include <unistd.h> // sleep
#include <stdlib.h> // atoi
#include <errno.h>
#include <signal.h>

// Needed by the "killer" thread
#include <pthread.h>
#include <semaphore.h>

#include <iostream>
#include <string> // needed by corba_option_parser

#ifdef ORB_MICO
#include <plugin_factory/plugin_factory.h>
#include <mico/CosNaming.h>
#elif defined (ORB_OMNIORB)
#include <plugin_factory/plugin_factory.hh>
#endif

#include <miscutils/corba_option_parser.h>

#include <miscutils/hostname.h> // my own version of getHostname()

using namespace DLADSPA;
using namespace std;

#include <plugin_factory/servant.h>


/******************************************************************************
 *      Handle the actual creation/destruction of the CORBA objects
 ******************************************************************************/

class FactoryManager
{
public:
    virtual ~FactoryManager ();
protected:
    void startup();
    void shutdown ();
    CosNaming::Name contextName;
    CosNaming::Name factoryName;
    CORBA::Object_ptr factory;
    virtual void startServant() = 0;
    virtual void stopServant () = 0;
    virtual void defineCorbaNames () = 0;
private:
    void registerFactory ();
    void deregisterFactory ();
};

class DLADSPAFactoryManager: public FactoryManager
{
public:
    DLADSPAFactoryManager ();
    ~DLADSPAFactoryManager ();
protected:
    void startServant ();
    void stopServant ();
    void defineCorbaNames ();
private:
    PluginFactoryImp servant;
};


/******************************************************************************
 *                            signal handling
 ******************************************************************************/

sigset_t killer_signals;

// This function will run in a separate thread to kill the
// ORB when a signal arrives
void*
killer (void* ptr)
{
    FactoryManager* builder = static_cast<FactoryManager*>(ptr);
    int i;
    sigwait (&killer_signals, &i);
    delete builder;
    CORBAOptionParser options;
    options.orb()->shutdown (false);
    return 0;
};

void
setupSignalHandling (void* ptr)
{
    // Mask the "killer" signals on this thread
    sigemptyset (&killer_signals);
    sigaddset (&killer_signals, SIGTERM);
    sigaddset (&killer_signals, SIGHUP);
    sigaddset (&killer_signals, SIGINT);
    sigaddset (&killer_signals, SIGQUIT);
    pthread_sigmask (SIG_BLOCK, &killer_signals, 0);

    // Prepare attributes for the "killer" thread
    pthread_attr_t thread_attributes;
    pthread_attr_init (&thread_attributes);
    pthread_attr_setdetachstate (&thread_attributes, PTHREAD_CREATE_DETACHED);

    // Create the "killer" thread
    pthread_t killer_thread;
    pthread_create (&killer_thread, &thread_attributes, killer, ptr);
};

/******************************************************************************
 *                                  MAIN
 ******************************************************************************/

int
main (int argc, char** argv)
{
    // This just starts the ORB
    CORBAOptionParser options (argc, argv);

    // Lock all address space in RAM
    mlockall(MCL_FUTURE);

    // Create the servant, register with nameserver
    FactoryManager* builder = new DLADSPAFactoryManager;

    // We setup the signal handler after creating the builder because
    // we need to delete it when receiving a signal.
    setupSignalHandling (static_cast<void*>(builder));

    // Start the ORB; if this returns, that's because the "killer" thread
    // was woken up; all objects have been cleaned and orb->shutdown()
    // has been called.
    options.orb()->run ();
}


/******************************************************************************
 *                  Implementation of FactoryManager
 ******************************************************************************/

void
FactoryManager::startup ()
{
    defineCorbaNames ();
    startServant ();
    registerFactory ();
};

void
FactoryManager::shutdown ()
{
    stopServant ();
    deregisterFactory ();  
};

FactoryManager::~FactoryManager() { };

void
FactoryManager::registerFactory ()
{
    CORBAOptionParser options;
    bool contextCreatedNow = false;
    CosNaming::NamingContext_var context;

    try {
        try {
            CORBA::Object_var obj = options.inc()->resolve (contextName);
            context = CosNaming::NamingContext::_narrow (obj);
        } catch (CosNaming::NamingContext::NotFound&) {
            // The context doesn't exist yet; let's create it.
            context = options.inc()->bind_new_context(contextName);
            contextCreatedNow = true;
        }
        // Let's register ourselves inside this context
//#ifdef notdef
        // FIXME!!!! This bit triggers a bug somewhere in MICO:
        // Compressor 0 from test plugin 1 doesn't work.
        try {
            // Remove any leftovers from a previous run
            context->unbind (factoryName);
        } catch (CosNaming::NamingContext::NotFound&) { };
//#endif
        try {
            context->bind (factoryName, factory);
        } catch (...) {
            // We couldn't register; weird... let's clean up and get out
            if (contextCreatedNow) {
                options.inc()->unbind (contextName);
                context->destroy();
            }
            throw;
        }

    } catch (...) {
       cerr << endl <<
        "    server.cc: Name server error; aborting..." << endl << endl;
        exit (1);
    }
};

void
FactoryManager::deregisterFactory ()
{
    CORBAOptionParser options;
    CORBA::Object_ptr obj = options.inc()->resolve (contextName);
    CosNaming::NamingContext_var context =
                                    CosNaming::NamingContext::_narrow (obj);
    CORBA::release (obj);
    context->unbind (factoryName);
    CosNaming::BindingList_var list;
    CosNaming::BindingIterator_var it;
    context->list (1, list, it);
    if (! CORBA::is_nil (it)) {
        it->destroy();
    }
    if (list->length() == 0) {
        // this context is empty; let's free it
        options.inc()->unbind (contextName);
        context->destroy();
    }
};

DLADSPAFactoryManager::DLADSPAFactoryManager ()
{
    startup();
};

DLADSPAFactoryManager::~DLADSPAFactoryManager ()
{
    shutdown ();
};

void
DLADSPAFactoryManager::startServant ()
{
    factory = servant._this ();
};

void
DLADSPAFactoryManager::stopServant ()
{
    // Maybe there's a simpler way to do this?
    CORBAOptionParser options;
    PortableServer::ObjectId* oid = options.poa()->reference_to_id (factory);
    options.poa()->deactivate_object (*oid);
    delete oid;
};

// This is actually a procedural style; we should have acessor methods
// that, if necessary, perform the steps here and return the adequate names.
void
DLADSPAFactoryManager::defineCorbaNames ()
{
    contextName.length (1);
    contextName[0].id = CORBA::string_dup ("DLADSPA");
    contextName[0].kind = CORBA::string_dup ("");

    factoryName.length (1);
    factoryName[0].id = CORBA::string_dup (DLowLat::getHostname().c_str());
    factoryName[0].kind = CORBA::string_dup ("pluginFactory");
};
